#include "sandboxed_api/sandbox2/syscall_defs.h"

#include <sys/types.h>

#include <array>
#include <cstdint>
#include <string>
#include <vector>

#include "absl/algorithm/container.h"
#include "absl/status/status.h"
#include "absl/status/statusor.h"
#include "absl/strings/escaping.h"
#include "absl/strings/str_cat.h"
#include "absl/strings/str_format.h"
#include "absl/strings/string_view.h"
#include "sandboxed_api/config.h"
#include "sandboxed_api/sandbox2/util.h"

namespace sandbox2 {

std::string SyscallTable::Entry::GetArgumentDescription(uint64_t value,
                                                        syscalls::ArgType type,
                                                        pid_t pid) {
  std::string ret = absl::StrFormat("%#x", value);
  switch (type) {
    case syscalls::kOct:
      absl::StrAppendFormat(&ret, " [\\0%o]", value);
      break;
    case syscalls::kPath:
      if (auto path_or = util::ReadCPathFromPid(pid, value); path_or.ok()) {
        absl::StrAppendFormat(&ret, " ['%s']",
                              absl::CHexEscape(path_or.value()));
      } else {
        absl::StrAppend(&ret, " [unreadable path]");
      }
      break;
    case syscalls::kInt:
      absl::StrAppendFormat(&ret, " [%d]", value);
      break;
    default:
      break;
  }
  return ret;
}

absl::string_view SyscallTable::GetName(int syscall) const {
  auto entry = GetEntry(syscall);
  if (!entry.ok()) {
    return "";
  }
  return entry->name;
}

namespace {

template <typename... ArgTypes>
constexpr SyscallTable::Entry MakeEntry(int nr, absl::string_view name,
                                        ArgTypes... arg_types) {
  static_assert(sizeof...(arg_types) <= syscalls::kMaxArgs,
                "Too many arguments for syscall");
  return {nr, name, sizeof...(arg_types), {arg_types...}};
}

struct UnknownArguments {};
constexpr SyscallTable::Entry MakeEntry(int nr, absl::string_view name,
                                        UnknownArguments) {
  return {nr,
          name,
          -1,
          {syscalls::kGen, syscalls::kGen, syscalls::kGen, syscalls::kGen,
           syscalls::kGen, syscalls::kGen}};
}

}  // namespace

absl::StatusOr<SyscallTable::Entry> SyscallTable::GetEntry(int syscall) const {
  auto it = absl::c_lower_bound(
      data_, syscall, [](const SyscallTable::Entry& entry, int syscall) {
        return entry.nr < syscall;
      });
  if (it == data_.end() || it->nr != syscall) {
    return absl::NotFoundError(absl::StrCat("Syscall not found: ", syscall));
  }
  return *it;
}

absl::StatusOr<SyscallTable::Entry> SyscallTable::GetEntry(
    absl::string_view name) const {
  // Note: There's no uniqueness guarantee of syscall names in the table, but
  // other than typos it's likely safe to assume uniqueness.
  auto filter = [name](const SyscallTable::Entry& entry) {
    return entry.name == name;
  };
  auto it = absl::c_find_if(data_, filter);
  if (it != data_.end()) {
    return *it;
  } else {
    return absl::NotFoundError(absl::StrCat("Name not found: ", name));
  }
}

std::vector<std::string> SyscallTable::GetArgumentsDescription(
    int syscall, const uint64_t values[], pid_t pid) const {
  static SyscallTable::Entry kInvalidEntry =
      MakeEntry(-1, "", UnknownArguments());
  auto entry = GetEntry(syscall);
  if (!entry.ok()) {
    entry = kInvalidEntry;
  }

  int num_args = entry->GetNumArgs();
  std::vector<std::string> rv;
  rv.reserve(num_args);
  for (int i = 0; i < num_args; ++i) {
    rv.push_back(SyscallTable::Entry::GetArgumentDescription(
        values[i], entry->arg_types[i], pid));
  }
  return rv;
}

namespace syscalls {
namespace {

// TODO(C++20) Use std::is_sorted
template <typename Container, typename Compare>
constexpr bool IsSorted(const Container& container, Compare comp) {
  auto it = std::begin(container);
  if (it == std::end(container)) {
    return true;
  }
  auto last = it;
  for (++it; it != std::end(container); ++it) {
    if (!comp(*last, *it)) {
      return false;
    }
    last = it;
  }
  return true;
}

// Syscall description table for Linux x86_64
constexpr std::array kSyscallDataX8664 = {
    // clang-format off
    MakeEntry(0, "read", kInt, kHex, kInt),
    MakeEntry(1, "write", kInt, kHex, kInt),
    MakeEntry(2, "open", kPath, kHex, kOct),
    MakeEntry(3, "close", kInt),
    MakeEntry(4, "stat", kPath, kGen),
    MakeEntry(5, "fstat", kInt, kHex),
    MakeEntry(6, "lstat", kPath, kGen),
    MakeEntry(7, "poll", kGen, kInt, kInt),
    MakeEntry(8, "lseek", kInt, kInt, kInt),
    MakeEntry(9, "mmap", kHex, kInt, kHex, kHex, kInt, kInt),
    MakeEntry(10, "mprotect", kHex, kInt, kHex),
    MakeEntry(11, "munmap", kHex, kInt),
    MakeEntry(12, "brk", kInt),
    MakeEntry(13, "rt_sigaction", kSignal, kHex, kHex, kInt),
    MakeEntry(14, "rt_sigprocmask", kInt, kHex, kHex, kInt),
    MakeEntry(15, "rt_sigreturn"),
    MakeEntry(16, "ioctl", kInt, kInt, kHex),
    MakeEntry(17, "pread64", kInt, kHex, kInt, kInt),
    MakeEntry(18, "pwrite64", kInt, kHex, kInt, kInt),
    MakeEntry(19, "readv", kInt, kHex, kInt),
    MakeEntry(20, "writev", kInt, kHex, kInt),
    MakeEntry(21, "access", kPath, kOct),
    MakeEntry(22, "pipe", kHex),
    MakeEntry(23, "select", kInt, kHex, kHex, kHex, kHex),
    MakeEntry(24, "sched_yield"),
    MakeEntry(25, "mremap", kHex, kInt, kInt, kInt, kHex),
    MakeEntry(26, "msync", kHex, kInt, kInt),
    MakeEntry(27, "mincore", kHex, kInt, kHex),
    MakeEntry(28, "madvise", kHex, kInt, kInt),
    MakeEntry(29, "shmget", kInt, kInt, kHex),
    MakeEntry(30, "shmat", kInt, kHex, kHex),
    MakeEntry(31, "shmctl", kInt, kInt, kHex),
    MakeEntry(32, "dup", kInt),
    MakeEntry(33, "dup2", kInt, kInt),
    MakeEntry(34, "pause"),
    MakeEntry(35, "nanosleep", kHex, kHex),
    MakeEntry(36, "getitimer", kInt, kHex),
    MakeEntry(37, "alarm", kInt),
    MakeEntry(38, "setitimer", kInt, kHex, kHex),
    MakeEntry(39, "getpid"),
    MakeEntry(40, "sendfile", kInt, kInt, kHex, kInt),
    MakeEntry(41, "socket", kAddressFamily, kInt, kInt),
    MakeEntry(42, "connect", kInt, kSockaddr, kInt),
    MakeEntry(43, "accept", kInt, kSockaddr, kHex),
    MakeEntry(44, "sendto", kInt, kHex, kInt, kHex, kSockaddr, kInt),
    MakeEntry(45, "recvfrom", kInt, kHex, kInt, kHex, kSockaddr, kHex),
    MakeEntry(46, "sendmsg", kInt, kSockmsghdr, kHex),
    MakeEntry(47, "recvmsg", kInt, kHex, kInt),
    MakeEntry(48, "shutdown", kInt, kInt),
    MakeEntry(49, "bind", kInt, kSockaddr, kInt),
    MakeEntry(50, "listen", kInt, kInt),
    MakeEntry(51, "getsockname", kInt, kSockaddr, kHex),
    MakeEntry(52, "getpeername", kInt, kSockaddr, kHex),
    MakeEntry(53, "socketpair", kAddressFamily, kInt, kInt, kHex),
    MakeEntry(54, "setsockopt", kInt, kInt, kInt, kHex, kHex),
    MakeEntry(55, "getsockopt", kInt, kInt, kInt, kHex, kInt),
    MakeEntry(56, "clone", kCloneFlag, kHex, kHex, kHex, kHex),
    MakeEntry(57, "fork"),
    MakeEntry(58, "vfork"),
    MakeEntry(59, "execve", kPath, kHex, kHex),
    MakeEntry(60, "exit", kInt),
    MakeEntry(61, "wait4", kInt, kHex, kHex, kHex),
    MakeEntry(62, "kill", kInt, kSignal),
    MakeEntry(63, "uname", kInt),
    MakeEntry(64, "semget", kInt, kInt, kHex),
    MakeEntry(65, "semop", kInt, kHex, kInt),
    MakeEntry(66, "semctl", kInt, kInt, kInt, kHex),
    MakeEntry(67, "shmdt", kHex),
    MakeEntry(68, "msgget", kInt, kHex),
    MakeEntry(69, "msgsnd", kInt, kHex, kInt, kHex),
    MakeEntry(70, "msgrcv", kInt, kHex, kInt, kInt, kHex),
    MakeEntry(71, "msgctl", kInt, kInt, kHex),
    MakeEntry(72, "fcntl", kInt, kInt, kHex),
    MakeEntry(73, "flock", kInt, kInt),
    MakeEntry(74, "fsync", kInt),
    MakeEntry(75, "fdatasync", kInt),
    MakeEntry(76, "truncate", kPath, kInt),
    MakeEntry(77, "ftruncate", kInt, kInt),
    MakeEntry(78, "getdents", kInt, kHex, kInt),
    MakeEntry(79, "getcwd", kHex, kInt),
    MakeEntry(80, "chdir", kPath),
    MakeEntry(81, "fchdir", kInt),
    MakeEntry(82, "rename", kPath, kPath),
    MakeEntry(83, "mkdir", kPath, kOct),
    MakeEntry(84, "rmdir", kPath),
    MakeEntry(85, "creat", kPath, kOct),
    MakeEntry(86, "link", kPath, kPath),
    MakeEntry(87, "unlink", kPath),
    MakeEntry(88, "symlink", kPath, kPath),
    MakeEntry(89, "readlink", kPath, kHex, kInt),
    MakeEntry(90, "chmod", kPath, kOct),
    MakeEntry(91, "fchmod", kInt, kOct),
    MakeEntry(92, "chown", kPath, kInt, kInt),
    MakeEntry(93, "fchown", kInt, kInt, kInt),
    MakeEntry(94, "lchown", kPath, kInt, kInt),
    MakeEntry(95, "umask", kHex),
    MakeEntry(96, "gettimeofday", kHex, kHex),
    MakeEntry(97, "getrlimit", kInt, kHex),
    MakeEntry(98, "getrusage", kInt, kHex),
    MakeEntry(99, "sysinfo", kHex),
    MakeEntry(100, "times", kHex),
    MakeEntry(101, "ptrace", kInt, kInt, kHex, kHex),
    MakeEntry(102, "getuid"),
    MakeEntry(103, "syslog", kInt, kHex, kInt),
    MakeEntry(104, "getgid"),
    MakeEntry(105, "setuid", kInt),
    MakeEntry(106, "setgid", kInt),
    MakeEntry(107, "geteuid"),
    MakeEntry(108, "getegid"),
    MakeEntry(109, "setpgid", kInt, kInt),
    MakeEntry(110, "getppid"),
    MakeEntry(111, "getpgrp"),
    MakeEntry(112, "setsid"),
    MakeEntry(113, "setreuid", kInt, kInt),
    MakeEntry(114, "setregid", kInt, kInt),
    MakeEntry(115, "getgroups", kInt, kHex),
    MakeEntry(116, "setgroups", kInt, kHex),
    MakeEntry(117, "setresuid", kInt, kInt, kInt),
    MakeEntry(118, "getresuid", kHex, kHex, kHex),
    MakeEntry(119, "setresgid", kInt, kInt, kInt),
    MakeEntry(120, "getresgid", kHex, kHex, kHex),
    MakeEntry(121, "getpgid", kInt),
    MakeEntry(122, "setfsuid", kInt),
    MakeEntry(123, "setfsgid", kInt),
    MakeEntry(124, "getsid", kInt),
    MakeEntry(125, "capget", kHex, kHex),
    MakeEntry(126, "capset", kHex, kHex),
    MakeEntry(127, "rt_sigpending", kHex, kInt),
    MakeEntry(128, "rt_sigtimedwait", kHex, kHex, kHex, kInt),
    MakeEntry(129, "rt_sigqueueinfo", kInt, kSignal, kHex),
    MakeEntry(130, "rt_sigsuspend", kHex, kInt),
    MakeEntry(131, "sigaltstack", kHex, kHex),
    MakeEntry(132, "utime", kPath, kHex),
    MakeEntry(133, "mknod", kPath, kOct, kHex),
    MakeEntry(134, "uselib", kPath),
    MakeEntry(135, "personality", kHex),
    MakeEntry(136, "ustat", kHex, kHex),
    MakeEntry(137, "statfs", kPath, kHex),
    MakeEntry(138, "fstatfs", kInt, kHex),
    MakeEntry(139, "sysfs", kInt, kInt, kInt),
    MakeEntry(140, "getpriority", kInt, kInt),
    MakeEntry(141, "setpriority", kInt, kInt, kInt),
    MakeEntry(142, "sched_setparam", kInt, kHex),
    MakeEntry(143, "sched_getparam", kInt, kHex),
    MakeEntry(144, "sched_setscheduler", kInt, kInt, kHex),
    MakeEntry(145, "sched_getscheduler", kInt),
    MakeEntry(146, "sched_get_priority_max", kInt),
    MakeEntry(147, "sched_get_priority_min", kInt),
    MakeEntry(148, "sched_rr_get_interval", kInt, kHex),
    MakeEntry(149, "mlock", kInt, kInt),
    MakeEntry(150, "munlock", kInt, kInt),
    MakeEntry(151, "mlockall", kHex),
    MakeEntry(152, "munlockall"),
    MakeEntry(153, "vhangup"),
    MakeEntry(154, "modify_ldt", kInt, kHex, kInt),
    MakeEntry(155, "pivot_root", kPath, kPath),
    MakeEntry(156, "_sysctl", kHex),
    MakeEntry(157, "prctl", kInt, kHex, kHex, kHex, kHex),
    MakeEntry(158, "arch_prctl", kInt, kHex),
    MakeEntry(159, "adjtimex", kHex),
    MakeEntry(160, "setrlimit", kInt, kHex),
    MakeEntry(161, "chroot", kPath),
    MakeEntry(162, "sync"),
    MakeEntry(163, "acct", kPath),
    MakeEntry(164, "settimeofday", kHex, kHex),
    MakeEntry(165, "mount", kPath, kPath, kString, kHex, kGen),
    MakeEntry(166, "umount2", kPath, kHex),
    MakeEntry(167, "swapon", kPath, kHex),
    MakeEntry(168, "swapoff", kPath),
    MakeEntry(169, "reboot", kInt, kHex, kHex, kGen),
    MakeEntry(170, "sethostname", kString, kInt),
    MakeEntry(171, "setdomainname", kString, kInt),
    MakeEntry(172, "iopl", kInt),
    MakeEntry(173, "ioperm", kInt, kInt, kInt),
    MakeEntry(174, "create_module", kString, kInt),
    MakeEntry(175, "init_module", kGen, kInt, kString),
    MakeEntry(176, "delete_module", kString, kHex),
    MakeEntry(177, "get_kernel_syms", kHex),
    MakeEntry(178, "query_module", kString, kInt, kGen, kInt, kGen),
    MakeEntry(179, "quotactl", kInt, kPath, kInt, kGen),
    MakeEntry(180, "nfsservctl", kInt, kGen, kGen),
    MakeEntry(181, "getpmsg", UnknownArguments()),
    MakeEntry(182, "putpmsg", UnknownArguments()),
    MakeEntry(183, "afs_syscall", UnknownArguments()),
    MakeEntry(184, "tuxcall", UnknownArguments()),
    MakeEntry(185, "security", UnknownArguments()),
    MakeEntry(186, "gettid"),
    MakeEntry(187, "readahead", kInt, kInt, kInt),
    MakeEntry(188, "setxattr", kPath, kString, kGen, kInt, kHex),
    MakeEntry(189, "lsetxattr", kPath, kString, kGen, kInt, kHex),
    MakeEntry(190, "fsetxattr", kInt, kString, kGen, kInt, kHex),
    MakeEntry(191, "getxattr", kPath, kString, kGen, kInt),
    MakeEntry(192, "lgetxattr", kPath, kString, kGen, kInt),
    MakeEntry(193, "fgetxattr", kInt, kString, kGen, kInt),
    MakeEntry(194, "listxattr", kPath, kGen, kInt),
    MakeEntry(195, "llistxattr", kPath, kGen, kInt),
    MakeEntry(196, "flistxattr", kInt, kGen, kInt),
    MakeEntry(197, "removexattr", kPath, kString),
    MakeEntry(198, "lremovexattr", kPath, kString),
    MakeEntry(199, "fremovexattr", kInt, kString),
    MakeEntry(200, "tkill", kInt, kSignal),
    MakeEntry(201, "time", kHex),
    MakeEntry(202, "futex", kGen, kInt, kInt, kGen, kGen, kInt),
    MakeEntry(203, "sched_setaffinity", kInt, kInt, kHex),
    MakeEntry(204, "sched_getaffinity", kInt, kInt, kHex),
    MakeEntry(205, "set_thread_area", kHex),
    MakeEntry(206, "io_setup", kInt, kHex),
    MakeEntry(207, "io_destroy", kInt),
    MakeEntry(208, "io_getevents", kInt, kInt, kInt, kHex, kHex),
    MakeEntry(209, "io_submit", kInt, kInt, kHex),
    MakeEntry(210, "io_cancel", kInt, kHex, kHex),
    MakeEntry(211, "get_thread_area", kHex),
    MakeEntry(212, "lookup_dcookie", kInt, kString, kInt),
    MakeEntry(213, "epoll_create", kInt),
    MakeEntry(214, "epoll_ctl_old", UnknownArguments()),
    MakeEntry(215, "epoll_wait_old", UnknownArguments()),
    MakeEntry(216, "remap_file_pages", kGen, kInt, kInt, kInt, kHex),
    MakeEntry(217, "getdents64", kInt, kHex, kInt),
    MakeEntry(218, "set_tid_address", kHex),
    MakeEntry(219, "restart_syscall"),
    MakeEntry(220, "semtimedop", kInt, kHex, kInt, kHex),
    MakeEntry(221, "fadvise64", kInt, kInt, kInt, kInt),
    MakeEntry(222, "timer_create", kInt, kHex, kHex),
    MakeEntry(223, "timer_settime", kInt, kHex, kHex, kHex),
    MakeEntry(224, "timer_gettime", kInt, kHex),
    MakeEntry(225, "timer_getoverrun", kInt),
    MakeEntry(226, "timer_delete", kInt),
    MakeEntry(227, "clock_settime", kInt, kHex),
    MakeEntry(228, "clock_gettime", kInt, kHex),
    MakeEntry(229, "clock_getres", kInt, kHex),
    MakeEntry(230, "clock_nanosleep", kInt, kHex, kHex, kHex),
    MakeEntry(231, "exit_group", kInt),
    MakeEntry(232, "epoll_wait", kInt, kHex, kInt, kInt),
    MakeEntry(233, "epoll_ctl", kInt, kInt, kInt, kHex),
    MakeEntry(234, "tgkill", kInt, kInt, kSignal),
    MakeEntry(235, "utimes", kPath, kHex),
    MakeEntry(236, "vserver", UnknownArguments()),
    MakeEntry(237, "mbind", kGen, kInt, kInt, kHex, kInt, kHex),
    MakeEntry(238, "set_mempolicy", kInt, kHex, kInt),
    MakeEntry(239, "get_mempolicy", kInt, kHex, kInt, kInt, kHex),
    MakeEntry(240, "mq_open", kString, kHex, kOct, kHex),
    MakeEntry(241, "mq_unlink", kString),
    MakeEntry(242, "mq_timedsend", kHex, kHex, kInt, kInt, kHex),
    MakeEntry(243, "mq_timedreceive", kHex, kHex, kInt, kHex, kHex),
    MakeEntry(244, "mq_notify", kHex, kHex),
    MakeEntry(245, "mq_getsetattr", kHex, kHex, kHex),
    MakeEntry(246, "kexec_load", kHex, kInt, kHex, kHex),
    MakeEntry(247, "waitid", kInt, kInt, kHex, kInt, kHex),
    MakeEntry(248, "add_key", kString, kString, kGen, kInt, kInt),
    MakeEntry(249, "request_key", kString, kString, kHex, kInt),
    MakeEntry(250, "keyctl", kInt, kInt, kInt, kInt, kInt),
    MakeEntry(251, "ioprio_set", kInt, kInt, kInt),
    MakeEntry(252, "ioprio_get", kInt, kInt),
    MakeEntry(253, "inotify_init"),
    MakeEntry(254, "inotify_add_watch", kInt, kPath, kHex),
    MakeEntry(255, "inotify_rm_watch", kInt, kInt),
    MakeEntry(256, "migrate_pages", kInt, kInt, kHex, kHex),
    MakeEntry(257, "openat", kInt, kPath, kHex, kOct),
    MakeEntry(258, "mkdirat", kInt, kPath, kOct),
    MakeEntry(259, "mknodat", kInt, kPath, kOct, kHex),
    MakeEntry(260, "fchownat", kInt, kPath, kInt, kInt, kHex),
    MakeEntry(261, "futimesat", kInt, kPath, kHex),
    MakeEntry(262, "newfstatat", kInt, kPath, kHex, kHex),
    MakeEntry(263, "unlinkat", kInt, kPath, kHex),
    MakeEntry(264, "renameat", kInt, kPath, kInt, kPath),
    MakeEntry(265, "linkat", kInt, kPath, kInt, kPath, kHex),
    MakeEntry(266, "symlinkat", kPath, kInt, kPath),
    MakeEntry(267, "readlinkat", kInt, kPath, kHex, kInt),
    MakeEntry(268, "fchmodat", kInt, kPath, kOct),
    MakeEntry(269, "faccessat", kInt, kPath, kInt, kHex),
    MakeEntry(270, "pselect6", kInt, kHex, kHex, kHex, kHex),
    MakeEntry(271, "ppoll", kHex, kInt, kHex, kHex, kInt),
    MakeEntry(272, "unshare", kHex),
    MakeEntry(273, "set_robust_list", kHex, kInt),
    MakeEntry(274, "get_robust_list", kInt, kHex, kHex),
    MakeEntry(275, "splice", kInt, kHex, kInt, kHex, kInt, kHex),
    MakeEntry(276, "tee", kInt, kInt, kInt, kHex),
    MakeEntry(277, "sync_file_range", kInt, kInt, kInt, kHex),
    MakeEntry(278, "vmsplice", kInt, kHex, kInt, kInt),
    MakeEntry(279, "move_pages", kInt, kInt, kHex, kHex, kHex, kHex),
    MakeEntry(280, "utimensat", kInt, kPath, kHex, kHex),
    MakeEntry(281, "epoll_pwait", kInt, kHex, kInt, kInt, kHex, kInt),
    MakeEntry(282, "signalfd", kInt, kHex, kHex),
    MakeEntry(283, "timerfd_create", kInt, kHex),
    MakeEntry(284, "eventfd", kInt),
    MakeEntry(285, "fallocate", kInt, kOct, kInt, kInt),
    MakeEntry(286, "timerfd_settime", kInt, kHex, kHex, kHex),
    MakeEntry(287, "timerfd_gettime", kInt, kHex),
    MakeEntry(288, "accept4", kInt, kHex, kHex, kInt),
    MakeEntry(289, "signalfd4", kInt, kHex, kHex, kHex),
    MakeEntry(290, "eventfd2", kInt, kHex),
    MakeEntry(291, "epoll_create1", kHex),
    MakeEntry(292, "dup3", kInt, kInt, kHex),
    MakeEntry(293, "pipe2", kHex, kHex),
    MakeEntry(294, "inotify_init1", kHex),
    MakeEntry(295, "preadv", kInt, kHex, kInt, kInt, kInt),
    MakeEntry(296, "pwritev", kInt, kHex, kInt, kInt, kInt),
    MakeEntry(297, "rt_tgsigqueueinfo", kInt, kInt, kInt, kHex),
    MakeEntry(298, "perf_event_open", kHex, kInt, kInt, kInt, kHex),
    MakeEntry(299, "recvmmsg", kInt, kHex, kInt, kHex, kHex),
    MakeEntry(300, "fanotify_init", kHex, kHex),
    MakeEntry(301, "fanotify_mark", kInt, kHex, kHex, kInt, kPath),
    MakeEntry(302, "prlimit64", kInt, kInt, kHex, kHex),
    MakeEntry(303, "name_to_handle_at", kInt, kPath, kHex, kHex, kHex),
    MakeEntry(304, "open_by_handle_at", kInt, kHex, kHex),
    MakeEntry(305, "clock_adjtime", kInt, kHex),
    MakeEntry(306, "syncfs", kInt),
    MakeEntry(307, "sendmmsg", kInt, kHex, kInt, kHex),
    MakeEntry(308, "setns", kInt, kHex),
    MakeEntry(309, "getcpu", kHex, kHex, kHex),
    MakeEntry(310, "process_vm_readv", kInt, kHex, kInt, kHex, kInt,
              kInt),
    MakeEntry(311, "process_vm_writev", kInt, kHex, kInt, kHex, kInt,
              kInt),
    MakeEntry(312, "kcmp", kInt, kInt, kInt, kInt, kInt),
    MakeEntry(313, "finit_module", kInt, kString, kHex),
    MakeEntry(314, "sched_setattr", kInt, kHex, kHex),
    MakeEntry(315, "sched_getattr", kInt, kHex, kInt, kHex),
    MakeEntry(316, "renameat2", kInt, kPath, kInt, kPath, kHex),
    MakeEntry(317, "seccomp", kInt, kHex, kHex),
    MakeEntry(318, "getrandom", kGen, kInt, kHex),
    MakeEntry(319, "memfd_create", kString, kHex),
    MakeEntry(320, "kexec_file_load", kInt, kInt, kInt, kString, kHex),
    MakeEntry(321, "bpf", kInt, kHex, kInt),
    MakeEntry(322, "execveat", kInt, kPath, kHex, kHex, kHex),
    MakeEntry(323, "userfaultfd", kHex),
    MakeEntry(324, "membarrier", kInt, kHex),
    MakeEntry(325, "mlock2", kHex, kInt, kHex),
    MakeEntry(326, "copy_file_range", kInt, kHex, kInt, kHex, kInt,
              kHex),
    MakeEntry(327, "preadv2", kInt, kHex, kInt, kInt, kInt, kHex),
    MakeEntry(328, "pwritev2", kInt, kHex, kInt, kInt, kInt, kHex),
    MakeEntry(329, "pkey_mprotect", kInt, kInt, kHex, kInt),
    MakeEntry(330, "pkey_alloc", kInt, kInt),
    MakeEntry(331, "pkey_free", kInt),
    MakeEntry(332, "statx", kInt, kPath, kHex, kHex, kHex),
    MakeEntry(333, "io_pgetevents", UnknownArguments()),
    MakeEntry(334, "rseq", kHex, kInt, kHex, kHex),
    MakeEntry(435, "clone3", kHex, kInt),
    MakeEntry(436, "close_range", kInt, kInt, kHex),
    MakeEntry(437, "openat2", kInt, kPath, kHex, kInt),
    MakeEntry(438, "pidfd_getfd", UnknownArguments()),
    MakeEntry(439, "faccessat2", kInt, kPath, kHex, kHex),
    MakeEntry(440, "process_madvise", UnknownArguments()),
    MakeEntry(441, "epoll_pwait2", UnknownArguments()),
    MakeEntry(442, "mount_setattr", UnknownArguments()),
    MakeEntry(443, "quotactl_fd", UnknownArguments()),
    MakeEntry(444, "landlock_create_ruleset", UnknownArguments()),
    MakeEntry(445, "landlock_add_rule", UnknownArguments()),
    MakeEntry(446, "landlock_restrict_self", UnknownArguments()),
    MakeEntry(447, "memfd_secret", UnknownArguments()),
    MakeEntry(448, "process_mrelease", UnknownArguments()),
    MakeEntry(449, "futex_waitv", UnknownArguments()),
    MakeEntry(450, "set_mempolicy_home_node", UnknownArguments()),
    MakeEntry(451, "cachestat", UnknownArguments()),
    MakeEntry(452, "fchmodat2", kInt, kPath, kHex, kHex),
    MakeEntry(453, "map_shadow_stack", UnknownArguments()),
    MakeEntry(454, "futex_wake", UnknownArguments()),
    MakeEntry(455, "futex_wait", UnknownArguments()),
    MakeEntry(456, "futex_requeue", UnknownArguments()),
    MakeEntry(457, "statmount", UnknownArguments()),
    MakeEntry(458, "listmount", UnknownArguments()),
    MakeEntry(459, "lsm_get_self_attr", UnknownArguments()),
    MakeEntry(460, "lsm_set_self_attr", UnknownArguments()),
    MakeEntry(461, "lsm_list_modules", UnknownArguments()),
    // clang-format on
};

static_assert(IsSorted(kSyscallDataX8664, SyscallTable::Entry::BySyscallNr),
              "Syscalls should be sorted");

constexpr std::array kSyscallDataX8632 = {
    // clang-format off
    MakeEntry(0, "restart_syscall"),
    MakeEntry(1, "exit", kInt),
    MakeEntry(2, "fork"),
    MakeEntry(3, "read", kInt, kHex, kInt),
    MakeEntry(4, "write", kInt, kHex, kInt),
    MakeEntry(5, "open", kPath, kHex, kOct),
    MakeEntry(6, "close", kInt),
    MakeEntry(7, "waitpid", kInt, kHex, kHex),
    MakeEntry(8, "creat", kPath, kOct),
    MakeEntry(9, "link", kPath, kPath),
    MakeEntry(10, "unlink", kPath),
    MakeEntry(11, "execve", kPath, kHex, kHex),
    MakeEntry(12, "chdir", kPath),
    MakeEntry(13, "time", kHex),
    MakeEntry(14, "mknod", kPath, kOct, kHex),
    MakeEntry(15, "chmod", kPath, kOct),
    MakeEntry(16, "lchown", kPath, kInt, kInt),
    MakeEntry(17, "break", UnknownArguments()),
    MakeEntry(18, "oldstat", UnknownArguments()),
    MakeEntry(19, "lseek", kInt, kInt, kInt),
    MakeEntry(20, "getpid"),
    MakeEntry(21, "mount", kPath, kPath, kString, kHex, kGen),
    MakeEntry(22, "umount", UnknownArguments()),
    MakeEntry(23, "setuid", kInt),
    MakeEntry(24, "getuid"),
    MakeEntry(25, "stime", UnknownArguments()),
    MakeEntry(26, "ptrace", kHex, kHex, kHex, kHex),
    MakeEntry(27, "alarm", kInt),
    MakeEntry(28, "oldfstat", UnknownArguments()),
    MakeEntry(29, "pause"),
    MakeEntry(30, "utime", kPath, kHex),
    MakeEntry(31, "stty", UnknownArguments()),
    MakeEntry(32, "gtty", UnknownArguments()),
    MakeEntry(33, "access", kPath, kHex),
    MakeEntry(34, "nice", UnknownArguments()),
    MakeEntry(35, "ftime", UnknownArguments()),
    MakeEntry(36, "sync"),
    MakeEntry(37, "kill", kInt, kSignal),
    MakeEntry(38, "rename", kPath, kPath),
    MakeEntry(39, "mkdir", kPath, kOct),
    MakeEntry(40, "rmdir", kPath),
    MakeEntry(41, "dup", kInt),
    MakeEntry(42, "pipe", kHex),
    MakeEntry(43, "times", kHex),
    MakeEntry(44, "prof", UnknownArguments()),
    MakeEntry(45, "brk", kInt),
    MakeEntry(46, "setgid", kInt),
    MakeEntry(47, "getgid"),
    MakeEntry(48, "signal", UnknownArguments()),
    MakeEntry(49, "geteuid"),
    MakeEntry(50, "getegid"),
    MakeEntry(51, "acct", kPath),
    MakeEntry(52, "umount2", kPath, kHex),
    MakeEntry(53, "lock", UnknownArguments()),
    MakeEntry(54, "ioctl", kInt, kInt, kHex),
    MakeEntry(55, "fcntl", kInt, kInt, kHex),
    MakeEntry(56, "mpx", UnknownArguments()),
    MakeEntry(57, "setpgid", kInt, kInt),
    MakeEntry(58, "ulimit", UnknownArguments()),
    MakeEntry(59, "oldolduname", UnknownArguments()),
    MakeEntry(60, "umask", kHex),
    MakeEntry(61, "chroot", kPath),
    MakeEntry(62, "ustat", kHex, kHex),
    MakeEntry(63, "dup2", kInt, kInt),
    MakeEntry(64, "getppid"),
    MakeEntry(65, "getpgrp"),
    MakeEntry(66, "setsid"),
    MakeEntry(67, "sigaction", UnknownArguments()),
    MakeEntry(68, "sgetmask", UnknownArguments()),
    MakeEntry(69, "ssetmask", UnknownArguments()),
    MakeEntry(70, "setreuid", kInt, kInt),
    MakeEntry(71, "setregid", kInt, kInt),
    MakeEntry(72, "sigsuspend", UnknownArguments()),
    MakeEntry(73, "sigpending", UnknownArguments()),
    MakeEntry(74, "sethostname", kString, kInt),
    MakeEntry(75, "setrlimit", kInt, kHex),
    MakeEntry(76, "getrlimit", kInt, kHex),
    MakeEntry(77, "getrusage", kInt, kHex),
    MakeEntry(78, "gettimeofday", kHex, kHex),
    MakeEntry(79, "settimeofday", kHex, kHex),
    MakeEntry(80, "getgroups", kInt, kHex),
    MakeEntry(81, "setgroups", kInt, kHex),
    MakeEntry(82, "select", kInt, kHex, kHex, kHex, kHex),
    MakeEntry(83, "symlink", kPath, kPath),
    MakeEntry(84, "oldlstat", UnknownArguments()),
    MakeEntry(85, "readlink", kPath, kHex, kInt),
    MakeEntry(86, "uselib", kPath),
    MakeEntry(87, "swapon", kPath, kHex),
    MakeEntry(88, "reboot", kInt, kHex, kHex, kGen),
    MakeEntry(89, "readdir", UnknownArguments()),
    MakeEntry(90, "mmap", kHex, kInt, kHex, kHex, kInt, kInt),
    MakeEntry(91, "munmap", kHex, kInt),
    MakeEntry(92, "truncate", kPath, kInt),
    MakeEntry(93, "ftruncate", kInt, kInt),
    MakeEntry(94, "fchmod", kInt, kOct),
    MakeEntry(95, "fchown", kInt, kInt, kInt),
    MakeEntry(96, "getpriority", kInt, kInt),
    MakeEntry(97, "setpriority", kInt, kInt, kInt),
    MakeEntry(98, "profil", UnknownArguments()),
    MakeEntry(99, "statfs", kPath, kHex),
    MakeEntry(100, "fstatfs", kInt, kHex),
    MakeEntry(101, "ioperm", kInt, kInt, kInt),
    MakeEntry(102, "socketcall", UnknownArguments()),
    MakeEntry(103, "syslog", kInt, kHex, kInt),
    MakeEntry(104, "setitimer", kInt, kHex, kHex),
    MakeEntry(105, "getitimer", kInt, kHex),
    MakeEntry(106, "stat", kPath, kHex),
    MakeEntry(107, "lstat", kPath, kHex),
    MakeEntry(108, "fstat", kInt, kHex),
    MakeEntry(109, "olduname", UnknownArguments()),
    MakeEntry(110, "iopl", kInt),
    MakeEntry(111, "vhangup"),
    MakeEntry(112, "idle", UnknownArguments()),
    MakeEntry(113, "vm86old", UnknownArguments()),
    MakeEntry(114, "wait4", kInt, kHex, kHex, kHex),
    MakeEntry(115, "swapoff", kPath),
    MakeEntry(116, "sysinfo", kHex),
    MakeEntry(117, "ipc", UnknownArguments()),
    MakeEntry(118, "fsync", kInt),
    MakeEntry(119, "sigreturn", UnknownArguments()),
    MakeEntry(120, "clone", kCloneFlag, kHex, kHex, kHex, kHex),
    MakeEntry(121, "setdomainname", kString, kInt),
    MakeEntry(122, "uname", kInt),
    MakeEntry(123, "modify_ldt", kInt, kHex, kInt),
    MakeEntry(124, "adjtimex", kHex),
    MakeEntry(125, "mprotect", kHex, kInt, kHex),
    MakeEntry(126, "sigprocmask", UnknownArguments()),
    MakeEntry(127, "create_module", kString, kInt),
    MakeEntry(128, "init_module", kGen, kInt, kString),
    MakeEntry(129, "delete_module", kString, kHex),
    MakeEntry(130, "get_kernel_syms", kHex),
    MakeEntry(131, "quotactl", kInt, kPath, kInt, kGen),
    MakeEntry(132, "getpgid", kInt),
    MakeEntry(133, "fchdir", kInt),
    MakeEntry(134, "bdflush", UnknownArguments()),
    MakeEntry(135, "sysfs", kInt, kInt, kInt),
    MakeEntry(136, "personality", kHex),
    MakeEntry(137, "afs_syscall", UnknownArguments()),
    MakeEntry(138, "setfsuid", kInt),
    MakeEntry(139, "setfsgid", kInt),
    MakeEntry(140, "_llseek", UnknownArguments()),
    MakeEntry(141, "getdents", kInt, kHex, kInt),
    MakeEntry(142, "_newselect", UnknownArguments()),
    MakeEntry(143, "flock", kInt, kInt),
    MakeEntry(144, "msync", kHex, kInt, kInt),
    MakeEntry(145, "readv", kInt, kHex, kInt),
    MakeEntry(146, "writev", kInt, kHex, kInt),
    MakeEntry(147, "getsid", kInt),
    MakeEntry(148, "fdatasync", kInt),
    MakeEntry(149, "_sysctl", kHex),
    MakeEntry(150, "mlock", kInt, kInt),
    MakeEntry(151, "munlock", kInt, kInt),
    MakeEntry(152, "mlockall", kHex),
    MakeEntry(153, "munlockall"),
    MakeEntry(154, "sched_setparam", kInt, kHex),
    MakeEntry(155, "sched_getparam", kInt, kHex),
    MakeEntry(156, "sched_setscheduler", kInt, kInt, kHex),
    MakeEntry(157, "sched_getscheduler", kInt),
    MakeEntry(158, "sched_yield"),
    MakeEntry(159, "sched_get_priority_max", kHex),
    MakeEntry(160, "sched_get_priority_min", kHex),
    MakeEntry(161, "sched_rr_get_interval", kInt, kHex),
    MakeEntry(162, "nanosleep", kHex, kHex),
    MakeEntry(163, "mremap", kHex, kInt, kInt, kInt, kHex),
    MakeEntry(164, "setresuid", kInt, kInt, kInt),
    MakeEntry(165, "getresuid", kHex, kHex, kHex),
    MakeEntry(166, "vm86", UnknownArguments()),
    MakeEntry(167, "query_module", kString, kInt, kGen, kInt, kGen),
    MakeEntry(168, "poll", kGen, kInt, kInt),
    MakeEntry(169, "nfsservctl", kInt, kGen, kGen),
    MakeEntry(170, "setresgid", kInt, kInt, kInt),
    MakeEntry(171, "getresgid", kHex, kHex, kHex),
    MakeEntry(172, "prctl", kInt, kHex, kHex, kHex, kHex),
    MakeEntry(173, "rt_sigreturn"),
    MakeEntry(174, "rt_sigaction", kSignal, kHex, kHex, kInt),
    MakeEntry(175, "rt_sigprocmask", kInt, kHex, kHex, kInt),
    MakeEntry(176, "rt_sigpending", kHex, kInt),
    MakeEntry(177, "rt_sigtimedwait", kHex, kHex, kHex, kInt),
    MakeEntry(178, "rt_sigqueueinfo", kInt, kSignal, kHex),
    MakeEntry(179, "rt_sigsuspend", kHex, kInt),
    MakeEntry(180, "pread64", kInt, kHex, kInt, kInt),
    MakeEntry(181, "pwrite64", kInt, kHex, kInt, kInt),
    MakeEntry(182, "chown", kPath, kInt, kInt),
    MakeEntry(183, "getcwd", kHex, kInt),
    MakeEntry(184, "capget", kHex, kHex),
    MakeEntry(185, "capset", kHex, kHex),
    MakeEntry(186, "sigaltstack", kHex, kHex),
    MakeEntry(187, "sendfile", kInt, kInt, kHex, kInt),
    MakeEntry(188, "getpmsg", UnknownArguments()),
    MakeEntry(189, "putpmsg", UnknownArguments()),
    MakeEntry(190, "vfork"),
    MakeEntry(191, "ugetrlimit", UnknownArguments()),
    MakeEntry(192, "mmap2", UnknownArguments()),
    MakeEntry(193, "truncate64", kPath, kHex),
    MakeEntry(194, "ftruncate64", kInt, kHex),
    MakeEntry(195, "stat64", kPath, kHex),
    MakeEntry(196, "lstat64", kPath, kHex),
    MakeEntry(197, "fstat64", kInt, kHex),
    MakeEntry(198, "lchown32", kPath, kInt, kInt),
    MakeEntry(199, "getuid32", UnknownArguments()),
    MakeEntry(200, "getgid32", UnknownArguments()),
    MakeEntry(201, "geteuid32", UnknownArguments()),
    MakeEntry(202, "getegid32", UnknownArguments()),
    MakeEntry(203, "setreuid32", UnknownArguments()),
    MakeEntry(204, "setregid32", UnknownArguments()),
    MakeEntry(205, "getgroups32", UnknownArguments()),
    MakeEntry(206, "setgroups32", UnknownArguments()),
    MakeEntry(207, "fchown32", UnknownArguments()),
    MakeEntry(208, "setresuid32", UnknownArguments()),
    MakeEntry(209, "getresuid32", UnknownArguments()),
    MakeEntry(210, "setresgid32", UnknownArguments()),
    MakeEntry(211, "getresgid32", UnknownArguments()),
    MakeEntry(212, "chown32", UnknownArguments()),
    MakeEntry(213, "setuid32", UnknownArguments()),
    MakeEntry(214, "setgid32", UnknownArguments()),
    MakeEntry(215, "setfsuid32", UnknownArguments()),
    MakeEntry(216, "setfsgid32", UnknownArguments()),
    MakeEntry(217, "pivot_root", kPath, kPath),
    MakeEntry(218, "mincore", kHex, kInt, kHex),
    MakeEntry(219, "madvise", kHex, kInt, kInt),
    MakeEntry(220, "getdents64", kInt, kHex, kInt),
    MakeEntry(221, "fcntl64", UnknownArguments()),
    MakeEntry(222, "unused1-222", UnknownArguments()),
    MakeEntry(223, "unused2-223", UnknownArguments()),
    MakeEntry(224, "gettid"),
    MakeEntry(225, "readahead", kInt, kInt, kInt),
    MakeEntry(226, "setxattr", kPath, kString, kGen, kInt, kHex),
    MakeEntry(227, "lsetxattr", kPath, kString, kGen, kInt, kHex),
    MakeEntry(228, "fsetxattr", kInt, kString, kGen, kInt, kHex),
    MakeEntry(229, "getxattr", kPath, kString, kGen, kInt),
    MakeEntry(230, "lgetxattr", kPath, kString, kGen, kInt),
    MakeEntry(231, "fgetxattr", kInt, kString, kGen, kInt),
    MakeEntry(232, "listxattr", kPath, kGen, kInt),
    MakeEntry(233, "llistxattr", kPath, kGen, kInt),
    MakeEntry(234, "flistxattr", kInt, kGen, kInt),
    MakeEntry(235, "removexattr", kPath, kString),
    MakeEntry(236, "lremovexattr", kPath, kString),
    MakeEntry(237, "fremovexattr", kInt, kString),
    MakeEntry(238, "tkill", kInt, kSignal),
    MakeEntry(239, "sendfile64", UnknownArguments()),
    MakeEntry(240, "futex", kGen, kInt, kInt, kGen, kGen, kInt),
    MakeEntry(241, "sched_setaffinity", kInt, kInt, kHex),
    MakeEntry(242, "sched_getaffinity", kInt, kInt, kHex),
    MakeEntry(243, "set_thread_area", kHex),
    MakeEntry(244, "get_thread_area", kHex),
    MakeEntry(245, "io_setup", kInt, kHex),
    MakeEntry(246, "io_destroy", kInt),
    MakeEntry(247, "io_getevents", kInt, kInt, kInt, kHex, kHex),
    MakeEntry(248, "io_submit", kInt, kInt, kHex),
    MakeEntry(249, "io_cancel", kInt, kHex, kHex),
    MakeEntry(250, "fadvise64", kInt, kInt, kInt, kInt),
    MakeEntry(251, "251-old_sys_set_zone_reclaim", UnknownArguments()),
    MakeEntry(252, "exit_group", kInt),
    MakeEntry(253, "lookup_dcookie", kInt, kString, kInt),
    MakeEntry(254, "epoll_create", kInt),
    MakeEntry(255, "epoll_ctl", kInt, kInt, kInt, kHex),
    MakeEntry(256, "epoll_wait", kInt, kHex, kInt, kInt),
    MakeEntry(257, "remap_file_pages", kGen, kInt, kInt, kInt, kHex),
    MakeEntry(258, "set_tid_address", kHex),
    MakeEntry(259, "timer_create", kInt, kHex, kHex),
    MakeEntry(260, "timer_settime", kInt, kHex, kHex, kHex),
    MakeEntry(261, "timer_gettime", kInt, kHex),
    MakeEntry(262, "timer_getoverrun", kInt),
    MakeEntry(263, "timer_delete", kInt),
    MakeEntry(264, "clock_settime", kInt, kHex),
    MakeEntry(265, "clock_gettime", kInt, kHex),
    MakeEntry(266, "clock_getres", kInt, kHex),
    MakeEntry(267, "clock_nanosleep", kInt, kHex, kHex, kHex),
    MakeEntry(268, "statfs64", UnknownArguments()),
    MakeEntry(269, "fstatfs64", UnknownArguments()),
    MakeEntry(270, "tgkill", kInt, kInt, kSignal),
    MakeEntry(271, "utimes", kPath, kHex),
    MakeEntry(272, "fadvise64_64", UnknownArguments()),
    MakeEntry(273, "vserver", UnknownArguments()),
    MakeEntry(274, "mbind", kGen, kInt, kInt, kHex, kInt, kHex),
    MakeEntry(275, "get_mempolicy", kInt, kHex, kInt, kInt, kHex),
    MakeEntry(276, "set_mempolicy", kInt, kHex, kInt),
    MakeEntry(277, "mq_open", kString, kHex, kOct, kHex),
    MakeEntry(278, "mq_unlink", kString),
    MakeEntry(279, "mq_timedsend", kHex, kHex, kInt, kInt, kHex),
    MakeEntry(280, "mq_timedreceive", kHex, kHex, kInt, kHex, kHex),
    MakeEntry(281, "mq_notify", kHex, kHex),
    MakeEntry(282, "mq_getsetattr", kHex, kHex, kHex),
    MakeEntry(283, "kexec_load", kHex, kInt, kHex, kHex),
    MakeEntry(284, "waitid", kInt, kInt, kHex, kInt, kHex),
    MakeEntry(285, "285-old_sys_setaltroot", UnknownArguments()),
    MakeEntry(286, "add_key", kString, kString, kGen, kInt, kInt),
    MakeEntry(287, "request_key", kString, kString, kHex, kInt),
    MakeEntry(288, "keyctl", kInt, kInt, kInt, kInt, kInt),
    MakeEntry(289, "ioprio_set", kInt, kInt, kInt),
    MakeEntry(290, "ioprio_get", kInt, kInt),
    MakeEntry(291, "inotify_init"),
    MakeEntry(292, "inotify_add_watch", kInt, kPath, kHex),
    MakeEntry(293, "inotify_rm_watch", kInt, kInt),
    MakeEntry(294, "migrate_pages", kInt, kInt, kHex, kHex),
    MakeEntry(295, "openat", kInt, kPath, kOct),
    MakeEntry(296, "mkdirat", kInt, kPath, kOct),
    MakeEntry(297, "mknodat", kInt, kPath, kOct, kHex),
    MakeEntry(298, "fchownat", kInt, kPath, kInt, kInt, kHex),
    MakeEntry(299, "futimesat", kInt, kPath, kHex),
    MakeEntry(300, "fstatat64", UnknownArguments()),
    MakeEntry(301, "unlinkat", kInt, kPath, kHex),
    MakeEntry(302, "renameat", kInt, kPath, kInt, kPath),
    MakeEntry(303, "linkat", kInt, kPath, kInt, kPath, kHex),
    MakeEntry(304, "symlinkat", kPath, kInt, kPath),
    MakeEntry(305, "readlinkat", kInt, kPath, kHex, kInt),
    MakeEntry(306, "fchmodat", kInt, kPath, kOct, kHex),
    MakeEntry(307, "faccessat", kInt, kPath, kHex, kHex),
    MakeEntry(308, "pselect6", kInt, kHex, kHex, kHex, kHex),
    MakeEntry(309, "ppoll", kHex, kInt, kHex, kHex, kInt),
    MakeEntry(310, "unshare", kHex),
    MakeEntry(311, "set_robust_list", kHex, kInt),
    MakeEntry(312, "get_robust_list", kInt, kHex, kHex),
    MakeEntry(313, "splice", kInt, kHex, kInt, kHex, kInt, kHex),
    MakeEntry(314, "sync_file_range", kInt, kInt, kInt, kHex),
    MakeEntry(315, "tee", kInt, kInt, kInt, kHex),
    MakeEntry(316, "vmsplice", kInt, kHex, kInt, kInt),
    MakeEntry(317, "move_pages", kInt, kInt, kHex, kHex, kHex, kHex),
    MakeEntry(318, "getcpu", kHex, kHex, kHex),
    MakeEntry(319, "epoll_pwait", kInt, kHex, kInt, kInt, kHex, kInt),
    MakeEntry(320, "utimensat", kInt, kPath, kHex, kHex),
    MakeEntry(321, "signalfd", kInt, kHex, kHex),
    MakeEntry(322, "timerfd_create", kInt, kHex),
    MakeEntry(323, "eventfd", kInt),
    MakeEntry(324, "fallocate", kInt, kOct, kInt, kInt),
    MakeEntry(325, "timerfd_settime", kInt, kHex, kHex, kHex),
    MakeEntry(326, "timerfd_gettime", kInt, kHex),
    MakeEntry(327, "signalfd4", kInt, kHex, kHex, kHex),
    MakeEntry(328, "eventfd2", kInt, kHex),
    MakeEntry(329, "epoll_create1", kHex),
    MakeEntry(330, "dup3", kInt, kInt, kHex),
    MakeEntry(331, "pipe2", kHex, kHex),
    MakeEntry(332, "inotify_init1", kHex),
    MakeEntry(333, "preadv", kInt, kHex, kInt, kInt, kInt),
    MakeEntry(334, "pwritev", kInt, kHex, kInt, kInt, kInt),
    MakeEntry(335, "rt_tgsigqueueinfo", kInt, kInt, kInt, kHex),
    MakeEntry(336, "perf_event_open", kHex, kInt, kInt, kInt, kHex),
    MakeEntry(337, "recvmmsg", kInt, kHex, kInt, kHex, kHex),
    MakeEntry(338, "fanotify_init", kHex, kHex),
    MakeEntry(339, "fanotify_mark", kInt, kHex, kHex, kInt, kPath),
    MakeEntry(340, "prlimit64", kInt, kInt, kHex, kHex),
    MakeEntry(341, "name_to_handle_at", kInt, kPath, kHex, kHex, kHex),
    MakeEntry(342, "open_by_handle_at", kInt, kHex, kHex),
    MakeEntry(343, "clock_adjtime", kInt, kHex),
    MakeEntry(344, "syncfs", kInt),
    MakeEntry(345, "sendmmsg", kInt, kHex, kInt, kHex),
    MakeEntry(346, "setns", kInt, kHex),
    MakeEntry(347, "process_vm_readv", kInt, kHex, kInt, kHex, kInt, kInt),
    MakeEntry(348, "process_vm_writev", kInt, kHex, kInt, kHex, kInt, kInt),
    MakeEntry(349, "kcmp", kInt, kInt, kInt, kInt, kInt),
    MakeEntry(350, "finit_module", kInt, kString, kHex),
    MakeEntry(351, "sched_setattr", kInt, kHex, kHex),
    MakeEntry(352, "sched_getattr", kInt, kHex, kInt, kHex),
    MakeEntry(353, "renameat2", kHex, kPath, kHex, kPath, kHex, kHex),
    MakeEntry(354, "seccomp", kInt, kHex, kHex),
    MakeEntry(355, "getrandom", kGen, kInt, kHex),
    MakeEntry(356, "memfd_create", kString, kHex),
    MakeEntry(357, "bpf", kInt, kHex, kInt),
    MakeEntry(435, "clone3", kHex, kInt),
    MakeEntry(436, "close_range", kInt, kInt, kHex),
    MakeEntry(437, "openat2", kInt, kPath, kHex, kInt),
    MakeEntry(438, "pidfd_getfd", UnknownArguments()),
    MakeEntry(439, "faccessat2", kInt, kPath, kHex, kHex),
    MakeEntry(440, "process_madvise", UnknownArguments()),
    MakeEntry(441, "epoll_pwait2", UnknownArguments()),
    MakeEntry(442, "mount_setattr", UnknownArguments()),
    MakeEntry(443, "quotactl_fd", UnknownArguments()),
    MakeEntry(444, "landlock_create_ruleset", UnknownArguments()),
    MakeEntry(445, "landlock_add_rule", UnknownArguments()),
    MakeEntry(446, "landlock_restrict_self", UnknownArguments()),
    MakeEntry(447, "memfd_secret", UnknownArguments()),
    MakeEntry(448, "process_mrelease", UnknownArguments()),
    MakeEntry(449, "futex_waitv", UnknownArguments()),
    MakeEntry(450, "set_mempolicy_home_node", UnknownArguments()),
    MakeEntry(451, "cachestat", UnknownArguments()),
    MakeEntry(452, "fchmodat2", kInt, kPath, kHex, kHex),
    MakeEntry(453, "map_shadow_stack", UnknownArguments()),
    MakeEntry(454, "futex_wake", UnknownArguments()),
    MakeEntry(455, "futex_wait", UnknownArguments()),
    MakeEntry(456, "futex_requeue", UnknownArguments()),
    MakeEntry(457, "statmount", UnknownArguments()),
    MakeEntry(458, "listmount", UnknownArguments()),
    MakeEntry(459, "lsm_get_self_attr", UnknownArguments()),
    MakeEntry(460, "lsm_set_self_attr", UnknownArguments()),
    MakeEntry(461, "lsm_list_modules", UnknownArguments()),
    // clang-format on
};

static_assert(IsSorted(kSyscallDataX8632, SyscallTable::Entry::BySyscallNr),
              "Syscalls should be sorted");

// http://lxr.free-electrons.com/source/arch/powerpc/include/uapi/asm/unistd.h
// Note: PPC64 syscalls can have up to 7 register arguments, but nobody is
// using the 7th argument - probably for x64 compatibility reasons.
constexpr std::array kSyscallDataPPC64LE = {
    // clang-format off
    MakeEntry(0, "restart_syscall"),
    MakeEntry(1, "exit", kInt),
    MakeEntry(2, "fork"),
    MakeEntry(3, "read", kInt, kHex, kInt),
    MakeEntry(4, "write", kInt, kHex, kInt),
    MakeEntry(5, "open", kPath, kHex, kOct),
    MakeEntry(6, "close", kInt),
    MakeEntry(7, "waitpid", kInt, kHex, kHex),
    MakeEntry(8, "creat", kPath, kOct),
    MakeEntry(9, "link", kPath, kPath),
    MakeEntry(10, "unlink", kPath),
    MakeEntry(11, "execve", kPath, kHex, kHex),
    MakeEntry(12, "chdir", kPath),
    MakeEntry(13, "time", kHex),
    MakeEntry(14, "mknod", kPath, kOct, kHex),
    MakeEntry(15, "chmod", kPath, kOct),
    MakeEntry(16, "lchown", kPath, kInt, kInt),
    MakeEntry(17, "break", UnknownArguments()),
    MakeEntry(18, "oldstat", UnknownArguments()),
    MakeEntry(19, "lseek", kInt, kInt, kInt),
    MakeEntry(20, "getpid"),
    MakeEntry(21, "mount", kPath, kPath, kString, kHex, kGen),
    MakeEntry(22, "umount", UnknownArguments()),
    MakeEntry(23, "setuid", kInt),
    MakeEntry(24, "getuid"),
    MakeEntry(25, "stime", UnknownArguments()),
    MakeEntry(26, "ptrace", kInt, kInt, kHex, kHex),
    MakeEntry(27, "alarm", kInt),
    MakeEntry(28, "oldfstat", UnknownArguments()),
    MakeEntry(29, "pause"),
    MakeEntry(30, "utime", kPath, kHex),
    MakeEntry(31, "stty", UnknownArguments()),
    MakeEntry(32, "gtty", UnknownArguments()),
    MakeEntry(33, "access", kPath, kOct),
    MakeEntry(34, "nice", UnknownArguments()),
    MakeEntry(35, "ftime", UnknownArguments()),
    MakeEntry(36, "sync"),
    MakeEntry(37, "kill", kInt, kSignal),
    MakeEntry(38, "rename", kPath, kPath),
    MakeEntry(39, "mkdir", kPath, kOct),
    MakeEntry(40, "rmdir", kPath),
    MakeEntry(41, "dup", kInt),
    MakeEntry(42, "pipe", kHex),
    MakeEntry(43, "times", kHex),
    MakeEntry(44, "prof", UnknownArguments()),
    MakeEntry(45, "brk", kInt),
    MakeEntry(46, "setgid", kInt),
    MakeEntry(47, "getgid"),
    MakeEntry(48, "signal", UnknownArguments()),
    MakeEntry(49, "geteuid"),
    MakeEntry(50, "getegid"),
    MakeEntry(51, "acct", kPath),
    MakeEntry(52, "umount2", kPath, kHex),
    MakeEntry(53, "lock", UnknownArguments()),
    MakeEntry(54, "ioctl", kInt, kInt, kHex),
    MakeEntry(55, "fcntl", kInt, kInt, kHex),
    MakeEntry(56, "mpx", UnknownArguments()),
    MakeEntry(57, "setpgid", kInt, kInt),
    MakeEntry(58, "ulimit", UnknownArguments()),
    MakeEntry(59, "oldolduname", UnknownArguments()),
    MakeEntry(60, "umask", kHex),
    MakeEntry(61, "chroot", kPath),
    MakeEntry(62, "ustat", kHex, kHex),
    MakeEntry(63, "dup2", kInt, kInt),
    MakeEntry(64, "getppid"),
    MakeEntry(65, "getpgrp"),
    MakeEntry(66, "setsid"),
    MakeEntry(67, "sigaction", UnknownArguments()),
    MakeEntry(68, "sgetmask", UnknownArguments()),
    MakeEntry(69, "ssetmask", UnknownArguments()),
    MakeEntry(70, "setreuid", kInt, kInt),
    MakeEntry(71, "setregid", kInt, kInt),
    MakeEntry(72, "sigsuspend", UnknownArguments()),
    MakeEntry(73, "sigpending", UnknownArguments()),
    MakeEntry(74, "sethostname", kString, kInt),
    MakeEntry(75, "setrlimit", kInt, kHex),
    MakeEntry(76, "getrlimit", kInt, kHex),
    MakeEntry(77, "getrusage", kInt, kHex),
    MakeEntry(78, "gettimeofday", kHex, kHex),
    MakeEntry(79, "settimeofday", kHex, kHex),
    MakeEntry(80, "getgroups", kInt, kHex),
    MakeEntry(81, "setgroups", kInt, kHex),
    MakeEntry(82, "select", kInt, kHex, kHex, kHex, kHex),
    MakeEntry(83, "symlink", kPath, kPath),
    MakeEntry(84, "oldlstat", UnknownArguments()),
    MakeEntry(85, "readlink", kPath, kHex, kInt),
    MakeEntry(86, "uselib", kPath),
    MakeEntry(87, "swapon", kPath, kHex),
    MakeEntry(88, "reboot", kInt, kHex, kHex, kGen),
    MakeEntry(89, "readdir", UnknownArguments()),
    MakeEntry(90, "mmap", kHex, kInt, kHex, kHex, kInt, kInt),
    MakeEntry(91, "munmap", kHex, kInt),
    MakeEntry(92, "truncate", kPath, kInt),
    MakeEntry(93, "ftruncate", kInt, kInt),
    MakeEntry(94, "fchmod", kInt, kOct),
    MakeEntry(95, "fchown", kInt, kInt, kInt),
    MakeEntry(96, "getpriority", kInt, kInt),
    MakeEntry(97, "setpriority", kInt, kInt, kInt),
    MakeEntry(98, "profil", UnknownArguments()),
    MakeEntry(99, "statfs", kPath, kHex),
    MakeEntry(100, "fstatfs", kInt, kHex),
    MakeEntry(101, "ioperm", kInt, kInt, kInt),
    MakeEntry(102, "socketcall", UnknownArguments()),
    MakeEntry(103, "syslog", kInt, kHex, kInt),
    MakeEntry(104, "setitimer", kInt, kHex, kHex),
    MakeEntry(105, "getitimer", kInt, kHex),
    MakeEntry(106, "stat", kPath, kGen),
    MakeEntry(107, "lstat", kPath, kGen),
    MakeEntry(108, "fstat", kInt, kHex),
    MakeEntry(109, "olduname", UnknownArguments()),
    MakeEntry(110, "iopl", kInt),
    MakeEntry(111, "vhangup"),
    MakeEntry(112, "idle", UnknownArguments()),
    MakeEntry(113, "vm86", UnknownArguments()),
    MakeEntry(114, "wait4", kInt, kHex, kHex, kHex),
    MakeEntry(115, "swapoff", kPath),
    MakeEntry(116, "sysinfo", kHex),
    MakeEntry(117, "ipc", UnknownArguments()),
    MakeEntry(118, "fsync", kInt),
    MakeEntry(119, "sigreturn", UnknownArguments()),
    MakeEntry(120, "clone", kCloneFlag, kHex, kHex, kHex, kHex),
    MakeEntry(121, "setdomainname", kString, kInt),
    MakeEntry(122, "uname", kInt),
    MakeEntry(123, "modify_ldt", kInt, kHex, kInt),
    MakeEntry(124, "adjtimex", kHex),
    MakeEntry(125, "mprotect", kHex, kInt, kHex),
    MakeEntry(126, "sigprocmask", UnknownArguments()),
    MakeEntry(127, "create_module", kString, kInt),
    MakeEntry(128, "init_module", kGen, kInt, kString),
    MakeEntry(129, "delete_module", kString, kHex),
    MakeEntry(130, "get_kernel_syms", kHex),
    MakeEntry(131, "quotactl", kInt, kPath, kInt, kGen),
    MakeEntry(132, "getpgid", kInt),
    MakeEntry(133, "fchdir", kInt),
    MakeEntry(134, "bdflush", UnknownArguments()),
    MakeEntry(135, "sysfs", kInt, kInt, kInt),
    MakeEntry(136, "personality", kHex),
    MakeEntry(137, "afs_syscall", UnknownArguments()),
    MakeEntry(138, "setfsuid", kInt),
    MakeEntry(139, "setfsgid", kInt),
    MakeEntry(140, "_llseek", UnknownArguments()),
    MakeEntry(141, "getdents", kInt, kHex, kInt),
    MakeEntry(142, "_newselect", UnknownArguments()),
    MakeEntry(143, "flock", kInt, kInt),
    MakeEntry(144, "msync", kHex, kInt, kInt),
    MakeEntry(145, "readv", kInt, kHex, kInt),
    MakeEntry(146, "writev", kInt, kHex, kInt),
    MakeEntry(147, "getsid", kInt),
    MakeEntry(148, "fdatasync", kInt),
    MakeEntry(149, "_sysctl", kHex),
    MakeEntry(150, "mlock", kInt, kInt),
    MakeEntry(151, "munlock", kInt, kInt),
    MakeEntry(152, "mlockall", kHex),
    MakeEntry(153, "munlockall"),
    MakeEntry(154, "sched_setparam", kInt, kHex),
    MakeEntry(155, "sched_getparam", kInt, kHex),
    MakeEntry(156, "sched_setscheduler", kInt, kInt, kHex),
    MakeEntry(157, "sched_getscheduler", kInt),
    MakeEntry(158, "sched_yield"),
    MakeEntry(159, "sched_get_priority_max", kInt),
    MakeEntry(160, "sched_get_priority_min", kInt),
    MakeEntry(161, "sched_rr_get_interval", kInt, kHex),
    MakeEntry(162, "nanosleep", kHex, kHex),
    MakeEntry(163, "mremap", kHex, kInt, kInt, kInt, kHex),
    MakeEntry(164, "setresuid", kInt, kInt, kInt),
    MakeEntry(165, "getresuid", kHex, kHex, kHex),
    MakeEntry(166, "query_module", kString, kInt, kGen, kInt, kGen),
    MakeEntry(167, "poll", kGen, kInt, kInt),
    MakeEntry(168, "nfsservctl", kInt, kGen, kGen),
    MakeEntry(169, "setresgid", kInt, kInt, kInt),
    MakeEntry(170, "getresgid", kHex, kHex, kHex),
    MakeEntry(171, "prctl", kInt, kHex, kHex, kHex, kHex),
    MakeEntry(172, "rt_sigreturn"),
    MakeEntry(173, "rt_sigaction", kSignal, kHex, kHex, kInt),
    MakeEntry(174, "rt_sigprocmask", kInt, kHex, kHex, kInt),
    MakeEntry(175, "rt_sigpending", kHex, kInt),
    MakeEntry(176, "rt_sigtimedwait", kHex, kHex, kHex, kInt),
    MakeEntry(177, "rt_sigqueueinfo", kInt, kSignal, kHex),
    MakeEntry(178, "rt_sigsuspend", kHex, kInt),
    MakeEntry(179, "pread64", kInt, kHex, kInt, kInt),
    MakeEntry(180, "pwrite64", kInt, kHex, kInt, kInt),
    MakeEntry(181, "chown", kPath, kInt, kInt),
    MakeEntry(182, "getcwd", kHex, kInt),
    MakeEntry(183, "capget", kHex, kHex),
    MakeEntry(184, "capset", kHex, kHex),
    MakeEntry(185, "sigaltstack", kHex, kHex),
    MakeEntry(186, "sendfile", kInt, kInt, kHex, kInt),
    MakeEntry(187, "getpmsg", UnknownArguments()),
    MakeEntry(188, "putpmsg", UnknownArguments()),
    MakeEntry(189, "vfork"),
    MakeEntry(190, "ugetrlimit", UnknownArguments()),
    MakeEntry(191, "readahead", kInt, kInt, kInt),
    MakeEntry(192, "mmap2", UnknownArguments()),
    MakeEntry(193, "truncate64", UnknownArguments()),
    MakeEntry(194, "ftruncate64", UnknownArguments()),
    MakeEntry(195, "stat64", UnknownArguments()),
    MakeEntry(196, "lstat64", UnknownArguments()),
    MakeEntry(197, "fstat64", UnknownArguments()),
    MakeEntry(198, "pciconfig_read", UnknownArguments()),
    MakeEntry(199, "pciconfig_write", UnknownArguments()),
    MakeEntry(200, "pciconfig_iobase", UnknownArguments()),
    MakeEntry(201, "multiplexer", UnknownArguments()),
    MakeEntry(202, "getdents64", kInt, kHex, kInt),
    MakeEntry(203, "pivot_root", kPath, kPath),
    MakeEntry(204, "fcntl64", UnknownArguments()),
    MakeEntry(205, "madvise", kHex, kInt, kInt),
    MakeEntry(206, "mincore", kHex, kInt, kHex),
    MakeEntry(207, "gettid"),
    MakeEntry(208, "tkill", kInt, kSignal),
    MakeEntry(209, "setxattr", kPath, kString, kGen, kInt, kHex),
    MakeEntry(210, "lsetxattr", kPath, kString, kGen, kInt, kHex),
    MakeEntry(211, "fsetxattr", kInt, kString, kGen, kInt, kHex),
    MakeEntry(212, "getxattr", kPath, kString, kGen, kInt),
    MakeEntry(213, "lgetxattr", kPath, kString, kGen, kInt),
    MakeEntry(214, "fgetxattr", kInt, kString, kGen, kInt),
    MakeEntry(215, "listxattr", kPath, kGen, kInt),
    MakeEntry(216, "llistxattr", kPath, kGen, kInt),
    MakeEntry(217, "flistxattr", kInt, kGen, kInt),
    MakeEntry(218, "removexattr", kPath, kString),
    MakeEntry(219, "lremovexattr", kPath, kString),
    MakeEntry(220, "fremovexattr", kInt, kString),
    MakeEntry(221, "futex", kGen, kInt, kInt, kGen, kGen, kInt),
    MakeEntry(222, "sched_setaffinity", kInt, kInt, kHex),
    MakeEntry(223, "sched_getaffinity", kInt, kInt, kHex),
    MakeEntry(225, "tuxcall", UnknownArguments()),
    MakeEntry(226, "sendfile64", UnknownArguments()),
    MakeEntry(227, "io_setup", kInt, kHex),
    MakeEntry(228, "io_destroy", kInt),
    MakeEntry(229, "io_getevents", kInt, kInt, kInt, kHex, kHex),
    MakeEntry(230, "io_submit", kInt, kInt, kHex),
    MakeEntry(231, "io_cancel", kInt, kHex, kHex),
    MakeEntry(232, "set_tid_address", kHex),
    MakeEntry(233, "fadvise64", kInt, kInt, kInt, kInt),
    MakeEntry(234, "exit_group", kInt),
    MakeEntry(235, "lookup_dcookie", kInt, kString, kInt),
    MakeEntry(236, "epoll_create", kInt),
    MakeEntry(237, "epoll_ctl", kInt, kInt, kInt, kHex),
    MakeEntry(238, "epoll_wait", kInt, kHex, kInt, kInt),
    MakeEntry(239, "remap_file_pages", kGen, kInt, kInt, kInt, kHex),
    MakeEntry(240, "timer_create", kInt, kHex, kHex),
    MakeEntry(241, "timer_settime", kInt, kHex, kHex, kHex),
    MakeEntry(242, "timer_gettime", kInt, kHex),
    MakeEntry(243, "timer_getoverrun", kInt),
    MakeEntry(244, "timer_delete", kInt),
    MakeEntry(245, "clock_settime", kInt, kHex),
    MakeEntry(246, "clock_gettime", kInt, kHex),
    MakeEntry(247, "clock_getres", kInt, kHex),
    MakeEntry(248, "clock_nanosleep", kInt, kHex, kHex, kHex),
    MakeEntry(249, "swapcontext", UnknownArguments()),
    MakeEntry(250, "tgkill", kInt, kInt, kSignal),
    MakeEntry(251, "utimes", kPath, kHex),
    MakeEntry(252, "statfs64", UnknownArguments()),
    MakeEntry(253, "fstatfs64", UnknownArguments()),
    MakeEntry(254, "fadvise64_64", UnknownArguments()),
    MakeEntry(255, "rtas", UnknownArguments()),
    MakeEntry(256, "sys_debug_setcontext", kHex, kHex, kHex, kHex, kHex, kHex),
    MakeEntry(258, "migrate_pages", kInt, kInt, kHex, kHex),
    MakeEntry(259, "mbind", kGen, kInt, kInt, kHex, kInt, kHex),
    MakeEntry(260, "get_mempolicy", kInt, kHex, kInt, kInt, kHex),
    MakeEntry(261, "set_mempolicy", kInt, kHex, kInt),
    MakeEntry(262, "mq_open", kString, kHex, kOct, kHex),
    MakeEntry(263, "mq_unlink", kString),
    MakeEntry(264, "mq_timedsend", kHex, kHex, kInt, kInt, kHex),
    MakeEntry(265, "mq_timedreceive", kHex, kHex, kInt, kHex, kHex),
    MakeEntry(266, "mq_notify", kHex, kHex),
    MakeEntry(267, "mq_getsetattr", kHex, kHex, kHex),
    MakeEntry(268, "kexec_load", kHex, kInt, kHex, kHex),
    MakeEntry(269, "add_key", kString, kString, kGen, kInt, kInt),
    MakeEntry(270, "request_key", kString, kString, kHex, kInt),
    MakeEntry(271, "keyctl", kInt, kInt, kInt, kInt, kInt),
    MakeEntry(272, "waitid", kInt, kInt, kHex, kInt, kHex),
    MakeEntry(273, "ioprio_set", kInt, kInt, kInt),
    MakeEntry(274, "ioprio_get", kInt, kInt),
    MakeEntry(275, "inotify_init"),
    MakeEntry(276, "inotify_add_watch", kInt, kPath, kHex),
    MakeEntry(277, "inotify_rm_watch", kInt, kInt),
    MakeEntry(278, "spu_run", UnknownArguments()),
    MakeEntry(279, "spu_create", UnknownArguments()),
    MakeEntry(280, "pselect6", kInt, kHex, kHex, kHex, kHex),
    MakeEntry(281, "ppoll", kHex, kInt, kHex, kHex, kInt),
    MakeEntry(282, "unshare", kHex),
    MakeEntry(283, "splice", kInt, kHex, kInt, kHex, kInt, kHex),
    MakeEntry(284, "tee", kInt, kInt, kInt, kHex),
    MakeEntry(285, "vmsplice", kInt, kHex, kInt, kInt),
    MakeEntry(286, "openat", kInt, kPath, kHex, kOct),
    MakeEntry(287, "mkdirat", kInt, kPath, kOct),
    MakeEntry(288, "mknodat", kInt, kPath, kOct, kHex),
    MakeEntry(289, "fchownat", kInt, kPath, kInt, kInt, kHex),
    MakeEntry(290, "futimesat", kInt, kPath, kHex),
    MakeEntry(291, "newfstatat", kInt, kPath, kHex, kHex),
    MakeEntry(292, "unlinkat", kInt, kPath, kHex),
    MakeEntry(293, "renameat", kInt, kPath, kInt, kPath),
    MakeEntry(294, "linkat", kInt, kPath, kInt, kPath, kHex),
    MakeEntry(295, "symlinkat", kPath, kInt, kPath),
    MakeEntry(296, "readlinkat", kInt, kPath, kHex, kInt),
    MakeEntry(297, "fchmodat", kInt, kPath, kOct),
    MakeEntry(298, "faccessat", kInt, kPath, kInt, kHex),
    MakeEntry(299, "get_robust_list", kInt, kHex, kHex),
    MakeEntry(300, "set_robust_list", kHex, kInt),
    MakeEntry(301, "move_pages", kInt, kInt, kHex, kHex, kHex, kHex),
    MakeEntry(302, "getcpu", kHex, kHex, kHex),
    MakeEntry(303, "epoll_pwait", kInt, kHex, kInt, kInt, kHex, kInt),
    MakeEntry(304, "utimensat", kInt, kPath, kHex, kHex),
    MakeEntry(305, "signalfd", kInt, kHex, kHex),
    MakeEntry(306, "timerfd_create", kInt, kHex),
    MakeEntry(307, "eventfd", kInt),
    MakeEntry(308, "sync_file_range2", UnknownArguments()),
    MakeEntry(309, "fallocate", kInt, kOct, kInt, kInt),
    MakeEntry(310, "subpage_prot", UnknownArguments()),
    MakeEntry(311, "timerfd_settime", kInt, kHex, kHex, kHex),
    MakeEntry(312, "timerfd_gettime", kInt, kHex),
    MakeEntry(313, "signalfd4", kInt, kHex, kHex, kHex),
    MakeEntry(314, "eventfd2", kInt, kHex),
    MakeEntry(315, "epoll_create1", kHex),
    MakeEntry(316, "dup3", kInt, kInt, kHex),
    MakeEntry(317, "pipe2", kHex, kHex),
    MakeEntry(318, "inotify_init1", kHex),
    MakeEntry(319, "perf_event_open", kHex, kInt, kInt, kInt, kHex),
    MakeEntry(320, "preadv", kInt, kHex, kInt, kInt, kInt),
    MakeEntry(321, "pwritev", kInt, kHex, kInt, kInt, kInt),
    MakeEntry(322, "rt_tgsigqueueinfo", kInt, kInt, kInt, kHex),
    MakeEntry(323, "fanotify_init", kHex, kHex, kInt, kGen, kGen, kGen),
    MakeEntry(324, "fanotify_mark", kInt, kHex, kHex, kInt, kPath),
    MakeEntry(325, "prlimit64", kInt, kInt, kHex, kHex),
    MakeEntry(326, "socket", kAddressFamily, kInt, kInt),
    MakeEntry(327, "bind", kInt, kSockaddr, kInt),
    MakeEntry(328, "connect", kInt, kSockaddr, kInt),
    MakeEntry(329, "listen", kInt, kInt),
    MakeEntry(330, "accept", kInt, kSockaddr, kHex),
    MakeEntry(331, "getsockname", kInt, kSockaddr, kHex),
    MakeEntry(332, "getpeername", kInt, kSockaddr, kHex),
    MakeEntry(333, "socketpair", kAddressFamily, kInt, kInt, kHex),
    MakeEntry(334, "send", UnknownArguments()),
    MakeEntry(335, "sendto", kInt, kGen, kInt, kHex, kSockaddr, kInt),
    MakeEntry(336, "recv", UnknownArguments()),
    MakeEntry(337, "recvfrom", kInt, kHex, kInt, kHex, kSockaddr, kHex),
    MakeEntry(338, "shutdown", kInt, kInt),
    MakeEntry(339, "setsockopt", kInt, kInt, kInt, kHex, kHex),
    MakeEntry(340, "getsockopt", kInt, kInt, kInt, kHex, kInt),
    MakeEntry(341, "sendmsg", kInt, kSockmsghdr, kHex),
    MakeEntry(342, "recvmsg", kInt, kHex, kInt),
    MakeEntry(343, "recvmmsg", kInt, kHex, kInt, kHex, kHex),
    MakeEntry(344, "accept4", kInt, kHex, kHex, kInt),
    MakeEntry(345, "name_to_handle_at", kInt, kPath, kHex, kHex, kHex),
    MakeEntry(346, "open_by_handle_at", kInt, kHex, kHex),
    MakeEntry(347, "clock_adjtime", kInt, kHex),
    MakeEntry(348, "syncfs", kInt),
    MakeEntry(349, "sendmmsg", kInt, kHex, kInt, kHex),
    MakeEntry(350, "setns", kInt, kHex),
    MakeEntry(351, "process_vm_readv", kInt, kHex, kInt, kHex, kInt, kInt),
    MakeEntry(352, "process_vm_writev", kInt, kHex, kInt, kHex, kInt, kInt),
    MakeEntry(353, "finit_module", kInt, kString, kHex),
    MakeEntry(354, "kcmp", kInt, kInt, kInt, kInt, kInt),
    MakeEntry(355, "sched_setattr", kInt, kHex, kHex),
    MakeEntry(356, "sched_getattr", kInt, kHex, kInt, kHex),
    MakeEntry(357, "renameat2", kInt, kPath, kInt, kPath, kHex),
    MakeEntry(358, "seccomp", kInt, kHex, kHex),
    MakeEntry(359, "getrandom", kGen, kInt, kHex),
    MakeEntry(360, "memfd_create", kString, kHex),
    MakeEntry(361, "bpf", kInt, kHex, kInt),
    MakeEntry(362, "execveat", kInt, kPath, kHex, kHex, kHex),
    MakeEntry(363, "switch_endian", UnknownArguments()),
    MakeEntry(364, "userfaultfd", kHex),
    MakeEntry(365, "membarrier", kInt, kHex),
    MakeEntry(378, "mlock2", kHex, kInt, kHex),
    MakeEntry(379, "copy_file_range", kInt, kHex, kInt, kHex, kInt, kHex),
    MakeEntry(380, "preadv2", kInt, kHex, kInt, kInt, kInt, kHex),
    MakeEntry(381, "pwritev2", kInt, kHex, kInt, kInt, kInt, kHex),
    MakeEntry(435, "clone3", kHex, kInt),
    MakeEntry(436, "close_range", kInt, kInt, kHex),
    MakeEntry(437, "openat2", kInt, kPath, kHex, kInt),
    MakeEntry(438, "pidfd_getfd", UnknownArguments()),
    MakeEntry(439, "faccessat2", kInt, kPath, kHex, kHex),
    MakeEntry(440, "process_madvise", UnknownArguments()),
    MakeEntry(441, "epoll_pwait2", UnknownArguments()),
    MakeEntry(442, "mount_setattr", UnknownArguments()),
    MakeEntry(443, "quotactl_fd", UnknownArguments()),
    MakeEntry(444, "landlock_create_ruleset", UnknownArguments()),
    MakeEntry(445, "landlock_add_rule", UnknownArguments()),
    MakeEntry(446, "landlock_restrict_self", UnknownArguments()),
    MakeEntry(448, "process_mrelease", UnknownArguments()),
    MakeEntry(449, "futex_waitv", UnknownArguments()),
    MakeEntry(450, "set_mempolicy_home_node", UnknownArguments()),
    MakeEntry(451, "cachestat", UnknownArguments()),
    MakeEntry(452, "fchmodat2", kInt, kPath, kHex, kHex),
    MakeEntry(453, "map_shadow_stack", UnknownArguments()),
    MakeEntry(454, "futex_wake", UnknownArguments()),
    MakeEntry(455, "futex_wait", UnknownArguments()),
    MakeEntry(456, "futex_requeue", UnknownArguments()),
    MakeEntry(457, "statmount", UnknownArguments()),
    MakeEntry(458, "listmount", UnknownArguments()),
    MakeEntry(459, "lsm_get_self_attr", UnknownArguments()),
    MakeEntry(460, "lsm_set_self_attr", UnknownArguments()),
    MakeEntry(461, "lsm_list_modules", UnknownArguments()),
    // clang-format on
};

static_assert(IsSorted(kSyscallDataPPC64LE, SyscallTable::Entry::BySyscallNr),
              "Syscalls should be sorted");

// https://github.com/torvalds/linux/blob/v5.8/include/uapi/asm-generic/unistd.h
constexpr std::array kSyscallDataArm64 = {
    // clang-format off
    MakeEntry(0, "io_setup", kInt, kHex),
    MakeEntry(1, "io_destroy", kInt),
    MakeEntry(2, "io_submit", kInt, kInt, kHex),
    MakeEntry(3, "io_cancel", kInt, kHex, kHex),
    MakeEntry(4, "io_getevents", kInt, kInt, kInt, kHex, kHex),
    MakeEntry(5, "setxattr", kPath, kString, kGen, kInt, kHex),
    MakeEntry(6, "lsetxattr", kPath, kString, kGen, kInt, kHex),
    MakeEntry(7, "fsetxattr", kInt, kString, kGen, kInt, kHex),
    MakeEntry(8, "getxattr", kPath, kString, kGen, kInt),
    MakeEntry(9, "lgetxattr", kPath, kString, kGen, kInt),
    MakeEntry(10, "fgetxattr", kInt, kString, kGen, kInt),
    MakeEntry(11, "listxattr", kPath, kGen, kInt),
    MakeEntry(12, "llistxattr", kPath, kGen, kInt),
    MakeEntry(13, "flistxattr", kInt, kGen, kInt),
    MakeEntry(14, "removexattr", kPath, kString),
    MakeEntry(15, "lremovexattr", kPath, kString),
    MakeEntry(16, "fremovexattr", kInt, kString),
    MakeEntry(17, "getcwd", kHex, kInt),
    MakeEntry(18, "lookup_dcookie", kInt, kString, kInt),
    MakeEntry(19, "eventfd2", kInt, kHex),
    MakeEntry(20, "epoll_create1", kHex),
    MakeEntry(21, "epoll_ctl", kInt, kInt, kInt, kHex),
    MakeEntry(22, "epoll_pwait", kInt, kHex, kInt, kInt, kHex, kInt),
    MakeEntry(23, "dup", kInt),
    MakeEntry(24, "dup3", kInt, kInt, kHex),
    MakeEntry(25, "fcntl", kInt, kInt, kHex),
    MakeEntry(26, "inotify_init1", kHex),
    MakeEntry(27, "inotify_add_watch", kInt, kPath, kHex),
    MakeEntry(28, "inotify_rm_watch", kInt, kInt),
    MakeEntry(29, "ioctl", kInt, kInt, kHex),
    MakeEntry(30, "ioprio_set", kInt, kInt, kInt),
    MakeEntry(31, "ioprio_get", kInt, kInt),
    MakeEntry(32, "flock", kInt, kInt),
    MakeEntry(33, "mknodat", kInt, kPath, kOct, kHex),
    MakeEntry(34, "mkdirat", kInt, kPath, kOct),
    MakeEntry(35, "unlinkat", kInt, kPath, kHex),
    MakeEntry(36, "symlinkat", kPath, kInt, kPath),
    MakeEntry(37, "linkat", kInt, kPath, kInt, kPath, kHex),
    MakeEntry(38, "renameat", kInt, kPath, kInt, kPath),
    MakeEntry(39, "umount2", kPath, kHex),
    MakeEntry(40, "mount", kPath, kPath, kString, kHex, kGen),
    MakeEntry(41, "pivot_root", kPath, kPath),
    MakeEntry(42, "nfsservctl", kInt, kGen, kGen),
    MakeEntry(43, "statfs", kPath, kHex),
    MakeEntry(44, "fstatfs", kInt, kHex),
    MakeEntry(45, "truncate", kPath, kInt),
    MakeEntry(46, "ftruncate", kInt, kInt),
    MakeEntry(47, "fallocate", kInt, kOct, kInt, kInt),
    MakeEntry(48, "faccessat", kInt, kPath, kHex, kHex),
    MakeEntry(49, "chdir", kPath),
    MakeEntry(50, "fchdir", kInt),
    MakeEntry(51, "chroot", kPath),
    MakeEntry(52, "fchmod", kInt, kOct),
    MakeEntry(53, "fchmodat", kInt, kPath, kOct, kHex),
    MakeEntry(54, "fchownat", kInt, kPath, kInt, kInt, kHex),
    MakeEntry(55, "fchown", kInt, kInt, kInt),
    MakeEntry(56, "openat", kGen, kPath, kOct, kHex, kGen, kGen),
    MakeEntry(57, "close", kInt),
    MakeEntry(58, "vhangup"),
    MakeEntry(59, "pipe2", kHex, kHex),
    MakeEntry(60, "quotactl", kInt, kPath, kInt, kGen),
    MakeEntry(61, "getdents64", kInt, kHex, kInt),
    MakeEntry(62, "lseek", kInt, kInt, kInt),
    MakeEntry(63, "read", kInt, kHex, kInt),
    MakeEntry(64, "write", kInt, kHex, kInt),
    MakeEntry(65, "readv", kInt, kHex, kInt),
    MakeEntry(66, "writev", kInt, kHex, kInt),
    MakeEntry(67, "pread64", kInt, kHex, kInt, kInt),
    MakeEntry(68, "pwrite64", kInt, kHex, kInt, kInt),
    MakeEntry(69, "preadv", kInt, kHex, kInt, kInt, kInt),
    MakeEntry(70, "pwritev", kInt, kHex, kInt, kInt, kInt),
    MakeEntry(71, "sendfile", kInt, kInt, kHex, kInt),
    MakeEntry(72, "pselect6", kInt, kHex, kHex, kHex, kHex),
    MakeEntry(73, "ppoll", kHex, kInt, kHex, kHex, kInt),
    MakeEntry(74, "signalfd4", kInt, kHex, kHex, kHex),
    MakeEntry(75, "vmsplice", kInt, kHex, kInt, kInt),
    MakeEntry(76, "splice", kInt, kHex, kInt, kHex, kInt, kHex),
    MakeEntry(77, "tee", kInt, kInt, kInt, kHex),
    MakeEntry(78, "readlinkat", kInt, kPath, kHex, kInt),
    MakeEntry(79, "newfstatat", kInt, kPath, kHex, kHex),
    MakeEntry(80, "fstat", kInt, kHex),
    MakeEntry(81, "sync"),
    MakeEntry(82, "fsync", kInt),
    MakeEntry(83, "fdatasync", kInt),
    MakeEntry(84, "sync_file_range", kInt, kInt, kInt, kHex),
    MakeEntry(85, "timerfd_create", kInt, kHex),
    MakeEntry(86, "timerfd_settime", kInt, kHex, kHex, kHex),
    MakeEntry(87, "timerfd_gettime", kInt, kHex),
    MakeEntry(88, "utimensat", kInt, kPath, kHex, kHex),
    MakeEntry(89, "acct", kPath),
    MakeEntry(90, "capget", kHex, kHex),
    MakeEntry(91, "capset", kHex, kHex),
    MakeEntry(92, "personality", kHex),
    MakeEntry(93, "exit", kInt),
    MakeEntry(94, "exit_group", kInt),
    MakeEntry(95, "waitid", kInt, kInt, kHex, kInt, kHex),
    MakeEntry(96, "set_tid_address", kHex),
    MakeEntry(97, "unshare", kHex),
    MakeEntry(98, "futex", kGen, kInt, kInt, kGen, kGen, kInt),
    MakeEntry(99, "set_robust_list", kHex, kInt),
    MakeEntry(100, "get_robust_list", kInt, kHex, kHex),
    MakeEntry(101, "nanosleep", kHex, kHex),
    MakeEntry(102, "getitimer", kInt, kHex),
    MakeEntry(103, "setitimer", kInt, kHex, kHex),
    MakeEntry(104, "kexec_load", kHex, kInt, kHex, kHex),
    MakeEntry(105, "init_module", kGen, kInt, kString),
    MakeEntry(106, "delete_module", kString, kHex),
    MakeEntry(107, "timer_create", kInt, kHex, kHex),
    MakeEntry(108, "timer_gettime", kInt, kHex),
    MakeEntry(109, "timer_getoverrun", kInt),
    MakeEntry(110, "timer_settime", kInt, kHex, kHex, kHex),
    MakeEntry(111, "timer_delete", kInt),
    MakeEntry(112, "clock_settime", kInt, kHex),
    MakeEntry(113, "clock_gettime", kInt, kHex),
    MakeEntry(114, "clock_getres", kInt, kHex),
    MakeEntry(115, "clock_nanosleep", kInt, kHex, kHex, kHex),
    MakeEntry(116, "syslog", kInt, kHex, kInt),
    MakeEntry(117, "ptrace", kHex, kHex, kHex, kHex),
    MakeEntry(118, "sched_setparam", kInt, kHex),
    MakeEntry(119, "sched_setscheduler", kInt, kInt, kHex),
    MakeEntry(120, "sched_getscheduler", kInt),
    MakeEntry(121, "sched_getparam", kInt, kHex),
    MakeEntry(122, "sched_setaffinity", kInt, kInt, kHex),
    MakeEntry(123, "sched_getaffinity", kInt, kInt, kHex),
    MakeEntry(124, "sched_yield"),
    MakeEntry(125, "sched_get_priority_max", kHex),
    MakeEntry(126, "sched_get_priority_min", kHex),
    MakeEntry(127, "sched_rr_get_interval", kInt, kHex),
    MakeEntry(128, "restart_syscall"),
    MakeEntry(129, "kill", kInt, kSignal),
    MakeEntry(130, "tkill", kInt, kSignal),
    MakeEntry(131, "tgkill", kInt, kInt, kSignal),
    MakeEntry(132, "sigaltstack", kHex, kHex),
    MakeEntry(133, "rt_sigsuspend", kHex, kInt),
    MakeEntry(134, "rt_sigaction", kSignal, kHex, kHex, kInt),
    MakeEntry(135, "rt_sigprocmask", kInt, kHex, kHex, kInt),
    MakeEntry(136, "rt_sigpending", kHex, kInt),
    MakeEntry(137, "rt_sigtimedwait", kHex, kHex, kHex, kInt),
    MakeEntry(138, "rt_sigqueueinfo", kInt, kSignal, kHex),
    MakeEntry(139, "rt_sigreturn"),
    MakeEntry(140, "setpriority", kInt, kInt, kInt),
    MakeEntry(141, "getpriority", kInt, kInt),
    MakeEntry(142, "reboot", kInt, kHex, kHex, kGen),
    MakeEntry(143, "setregid", kInt, kInt),
    MakeEntry(144, "setgid", kInt),
    MakeEntry(145, "setreuid", kInt, kInt),
    MakeEntry(146, "setuid", kInt),
    MakeEntry(147, "setresuid", kInt, kInt, kInt),
    MakeEntry(148, "getresuid", kHex, kHex, kHex),
    MakeEntry(149, "setresgid", kInt, kInt, kInt),
    MakeEntry(150, "getresgid", kHex, kHex, kHex),
    MakeEntry(151, "setfsuid", kInt),
    MakeEntry(152, "setfsgid", kInt),
    MakeEntry(153, "times", kHex),
    MakeEntry(154, "setpgid", kInt, kInt),
    MakeEntry(155, "getpgid", kInt),
    MakeEntry(156, "getsid", kInt),
    MakeEntry(157, "setsid"),
    MakeEntry(158, "getgroups", kInt, kHex),
    MakeEntry(159, "setgroups", kInt, kHex),
    MakeEntry(160, "uname", kInt),
    MakeEntry(161, "sethostname", kString, kInt),
    MakeEntry(162, "setdomainname", kString, kInt),
    MakeEntry(163, "getrlimit", kInt, kHex),
    MakeEntry(164, "setrlimit", kInt, kHex),
    MakeEntry(165, "getrusage", kInt, kHex),
    MakeEntry(166, "umask", kHex),
    MakeEntry(167, "prctl", kInt, kHex, kHex, kHex, kHex),
    MakeEntry(168, "getcpu", kHex, kHex, kHex),
    MakeEntry(169, "gettimeofday", kHex, kHex),
    MakeEntry(170, "settimeofday", kHex, kHex),
    MakeEntry(171, "adjtimex", kHex),
    MakeEntry(172, "getpid"),
    MakeEntry(173, "getppid"),
    MakeEntry(174, "getuid"),
    MakeEntry(175, "geteuid"),
    MakeEntry(176, "getgid"),
    MakeEntry(177, "getegid"),
    MakeEntry(178, "gettid"),
    MakeEntry(179, "sysinfo", kHex),
    MakeEntry(180, "mq_open", kString, kHex, kOct, kHex),
    MakeEntry(181, "mq_unlink", kString),
    MakeEntry(182, "mq_timedsend", kHex, kHex, kInt, kInt, kHex),
    MakeEntry(183, "mq_timedreceive", kHex, kHex, kInt, kHex, kHex),
    MakeEntry(184, "mq_notify", kHex, kHex),
    MakeEntry(185, "mq_getsetattr", kHex, kHex, kHex),
    MakeEntry(186, "msgget", kInt, kHex),
    MakeEntry(187, "msgctl", kInt, kInt, kHex),
    MakeEntry(188, "msgrcv", kInt, kHex, kInt, kInt, kHex),
    MakeEntry(189, "msgsnd", kInt, kHex, kInt, kHex),
    MakeEntry(190, "semget", kInt, kInt, kHex),
    MakeEntry(191, "semctl", kInt, kInt, kInt, kHex),
    MakeEntry(192, "semtimedop", kInt, kHex, kInt, kHex),
    MakeEntry(193, "semop", kInt, kHex, kInt),
    MakeEntry(194, "shmget", kInt, kInt, kHex),
    MakeEntry(195, "shmctl", kInt, kInt, kHex),
    MakeEntry(196, "shmat", kInt, kHex, kHex),
    MakeEntry(197, "shmdt", kHex),
    MakeEntry(198, "socket", kAddressFamily, kInt, kInt),
    MakeEntry(199, "socketpair", kAddressFamily, kInt, kInt, kHex),
    MakeEntry(200, "bind", kInt, kSockaddr, kInt),
    MakeEntry(201, "listen", kInt, kInt),
    MakeEntry(202, "accept", kInt, kSockaddr, kHex),
    MakeEntry(203, "connect", kInt, kSockaddr, kInt),
    MakeEntry(204, "getsockname", kInt, kSockaddr, kHex),
    MakeEntry(205, "getpeername", kInt, kSockaddr, kHex),
    MakeEntry(206, "sendto", kInt, kGen, kInt, kHex, kSockaddr, kInt),
    MakeEntry(207, "recvfrom", kInt, kHex, kInt, kHex, kSockaddr, kHex),
    MakeEntry(208, "setsockopt", kInt, kInt, kInt, kHex, kHex),
    MakeEntry(209, "getsockopt", kInt, kInt, kInt, kHex, kInt),
    MakeEntry(210, "shutdown", kInt, kInt),
    MakeEntry(211, "sendmsg", kInt, kSockmsghdr, kHex),
    MakeEntry(212, "recvmsg", kInt, kHex, kInt),
    MakeEntry(213, "readahead", kInt, kInt, kInt),
    MakeEntry(214, "brk", kInt),
    MakeEntry(215, "munmap", kHex, kInt),
    MakeEntry(216, "mremap", kHex, kInt, kInt, kInt, kHex),
    MakeEntry(217, "add_key", kString, kString, kGen, kInt, kInt),
    MakeEntry(218, "request_key", kString, kString, kHex, kInt),
    MakeEntry(219, "keyctl", kInt, kInt, kInt, kInt, kInt),
    MakeEntry(220, "clone", kCloneFlag, kHex, kHex, kHex, kHex),
    MakeEntry(221, "execve", kPath, kHex, kHex),
    MakeEntry(222, "mmap", kHex, kInt, kHex, kHex, kInt, kInt),
    MakeEntry(223, "fadvise64", kInt, kInt, kInt, kInt),
    MakeEntry(224, "swapon", kPath, kHex),
    MakeEntry(225, "swapoff", kPath),
    MakeEntry(226, "mprotect", kHex, kInt, kHex),
    MakeEntry(227, "msync", kHex, kInt, kInt),
    MakeEntry(228, "mlock", kInt, kInt),
    MakeEntry(229, "munlock", kInt, kInt),
    MakeEntry(230, "mlockall", kHex),
    MakeEntry(231, "munlockall"),
    MakeEntry(232, "mincore", kHex, kInt, kHex),
    MakeEntry(233, "madvise", kHex, kInt, kInt),
    MakeEntry(234, "remap_file_pages", kGen, kInt, kInt, kInt, kHex),
    MakeEntry(235, "mbind", kGen, kInt, kInt, kHex, kInt, kHex),
    MakeEntry(236, "get_mempolicy", kInt, kHex, kInt, kInt, kHex),
    MakeEntry(237, "set_mempolicy", kInt, kHex, kInt),
    MakeEntry(238, "migrate_pages", kInt, kInt, kHex, kHex),
    MakeEntry(239, "move_pages", kInt, kInt, kHex, kHex, kHex, kHex),
    MakeEntry(240, "rt_tgsigqueueinfo", kInt, kInt, kInt, kHex),
    MakeEntry(241, "perf_event_open", kHex, kInt, kInt, kInt, kHex),
    MakeEntry(242, "accept4", kInt, kHex, kHex, kInt),
    MakeEntry(243, "recvmmsg", kInt, kHex, kInt, kHex, kHex),
    MakeEntry(260, "wait4", kInt, kHex, kHex, kHex),
    MakeEntry(261, "prlimit64", kInt, kInt, kHex, kHex),
    MakeEntry(262, "fanotify_init", kHex, kHex),
    MakeEntry(263, "fanotify_mark", kInt, kHex, kHex, kInt, kPath),
    MakeEntry(264, "name_to_handle_at", kInt, kPath, kHex, kHex, kHex),
    MakeEntry(265, "open_by_handle_at", kInt, kHex, kHex),
    MakeEntry(266, "clock_adjtime", kInt, kHex),
    MakeEntry(267, "syncfs", kInt),
    MakeEntry(268, "setns", kInt, kHex),
    MakeEntry(269, "sendmmsg", kInt, kHex, kInt, kHex),
    MakeEntry(270, "process_vm_readv", kInt, kHex, kInt, kHex, kInt, kInt),
    MakeEntry(271, "process_vm_writev", kInt, kHex, kInt, kHex, kInt, kInt),
    MakeEntry(272, "kcmp", kInt, kInt, kInt, kInt, kInt),
    MakeEntry(273, "finit_module", kInt, kString, kHex),
    MakeEntry(274, "sched_setattr", kInt, kHex, kHex),
    MakeEntry(275, "sched_getattr", kInt, kHex, kInt, kHex),
    MakeEntry(276, "renameat2", kGen, kPath, kGen, kPath, kGen, kGen),
    MakeEntry(277, "seccomp", kInt, kHex, kHex),
    MakeEntry(278, "getrandom", kGen, kInt, kHex),
    MakeEntry(279, "memfd_create", kString, kHex),
    MakeEntry(280, "bpf", kInt, kHex, kInt),
    MakeEntry(281, "execveat", kInt, kPath, kHex, kHex, kHex),
    MakeEntry(282, "userfaultfd", kHex),
    MakeEntry(283, "membarrier", kInt, kHex),
    MakeEntry(284, "mlock2", kHex, kInt, kHex),
    MakeEntry(285, "copy_file_range", kInt, kHex, kInt, kHex, kInt, kHex),
    MakeEntry(286, "preadv2", kInt, kHex, kInt, kInt, kInt, kHex),
    MakeEntry(287, "pwritev2", kInt, kHex, kInt, kInt, kInt, kHex),
    MakeEntry(435, "clone3", kHex, kInt),
    MakeEntry(436, "close_range", kInt, kInt, kHex),
    MakeEntry(437, "openat2", kInt, kPath, kHex, kInt),
    MakeEntry(438, "pidfd_getfd", UnknownArguments()),
    MakeEntry(439, "faccessat2", kInt, kPath, kHex, kHex),
    MakeEntry(440, "process_madvise", UnknownArguments()),
    MakeEntry(441, "epoll_pwait2", UnknownArguments()),
    MakeEntry(442, "mount_setattr", UnknownArguments()),
    MakeEntry(443, "quotactl_fd", UnknownArguments()),
    MakeEntry(444, "landlock_create_ruleset", UnknownArguments()),
    MakeEntry(445, "landlock_add_rule", UnknownArguments()),
    MakeEntry(446, "landlock_restrict_self", UnknownArguments()),
    MakeEntry(448, "process_mrelease", UnknownArguments()),
    MakeEntry(449, "futex_waitv", UnknownArguments()),
    MakeEntry(450, "set_mempolicy_home_node", UnknownArguments()),
    MakeEntry(451, "cachestat", UnknownArguments()),
    MakeEntry(452, "fchmodat2", kInt, kPath, kHex, kHex),
    MakeEntry(453, "map_shadow_stack", UnknownArguments()),
    MakeEntry(454, "futex_wake", UnknownArguments()),
    MakeEntry(455, "futex_wait", UnknownArguments()),
    MakeEntry(456, "futex_requeue", UnknownArguments()),
    MakeEntry(457, "statmount", UnknownArguments()),
    MakeEntry(458, "listmount", UnknownArguments()),
    MakeEntry(459, "lsm_get_self_attr", UnknownArguments()),
    MakeEntry(460, "lsm_set_self_attr", UnknownArguments()),
    MakeEntry(461, "lsm_list_modules", UnknownArguments()),
    // clang-format on
};

static_assert(IsSorted(kSyscallDataArm64, SyscallTable::Entry::BySyscallNr),
              "Syscalls should be sorted");

constexpr std::array kSyscallDataArm32 = {
    // clang-format off
    MakeEntry(0, "restart_syscall"),
    MakeEntry(1, "exit", kHex, kHex, kHex, kHex),
    MakeEntry(2, "fork"),
    MakeEntry(3, "read", kInt, kHex, kInt),
    MakeEntry(4, "write", kHex, kHex, kHex, kHex),
    MakeEntry(5, "open", kPath, kHex, kOct),
    MakeEntry(6, "close", kHex, kHex, kHex, kHex),
    MakeEntry(8, "creat", kPath, kHex, kHex, kHex),
    MakeEntry(9, "link", kPath, kPath),
    MakeEntry(10, "unlink", kPath),
    MakeEntry(11, "execve", kPath, kHex, kHex),
    MakeEntry(12, "chdir", kPath),
    MakeEntry(14, "mknod", kPath, kOct, kHex),
    MakeEntry(15, "chmod", kPath, kOct),
    MakeEntry(16, "lchown", kPath, kInt, kInt),
    MakeEntry(19, "lseek", kInt, kInt, kInt),
    MakeEntry(20, "getpid"),
    MakeEntry(21, "mount", kPath, kPath, kString, kHex, kGen),
    MakeEntry(23, "setuid", kInt),
    MakeEntry(24, "getuid"),
    MakeEntry(26, "ptrace", kHex, kHex, kHex, kHex),
    MakeEntry(29, "pause"),
    MakeEntry(33, "access", kPath, kHex),
    MakeEntry(34, "nice", kHex, kHex, kHex, kHex),
    MakeEntry(36, "sync"),
    MakeEntry(37, "kill", kHex, kHex, kHex, kHex),
    MakeEntry(38, "rename", kPath, kPath),
    MakeEntry(39, "mkdir", kPath, kHex, kHex, kHex),
    MakeEntry(40, "rmdir", kHex, kHex, kHex, kHex),
    MakeEntry(41, "dup", kInt),
    MakeEntry(42, "pipe", kHex),
    MakeEntry(43, "times", kHex),
    MakeEntry(45, "brk", kHex),
    MakeEntry(46, "setgid", kInt),
    MakeEntry(47, "getgid"),
    MakeEntry(49, "geteuid"),
    MakeEntry(50, "getegid"),
    MakeEntry(51, "acct", kHex, kHex, kHex, kHex),
    MakeEntry(52, "umount2", kHex, kHex, kHex, kHex),
    MakeEntry(54, "ioctl", kInt, kInt, kHex),
    MakeEntry(55, "fcntl", kInt, kInt, kHex),
    MakeEntry(57, "setpgid", kInt, kInt),
    MakeEntry(60, "umask", kHex),
    MakeEntry(61, "chroot", kHex, kHex, kHex, kHex),
    MakeEntry(62, "ustat", kHex, kHex),
    MakeEntry(63, "dup2", kGen, kGen),
    MakeEntry(64, "getppid"),
    MakeEntry(65, "getpgrp"),
    MakeEntry(66, "setsid"),
    MakeEntry(67, "sigaction", kHex, kHex, kHex, kHex),
    MakeEntry(70, "setreuid", kInt, kInt),
    MakeEntry(71, "setregid", kInt, kInt),
    MakeEntry(72, "sigsuspend", kHex, kHex, kHex, kHex),
    MakeEntry(73, "sigpending", kHex, kHex, kHex, kHex),
    MakeEntry(74, "sethostname", kString, kInt),
    MakeEntry(75, "setrlimit", kInt, kHex),
    MakeEntry(77, "getrusage", kInt, kHex),
    MakeEntry(78, "gettimeofday", kHex, kHex),
    MakeEntry(79, "settimeofday", kHex, kHex),
    MakeEntry(80, "getgroups", kInt, kHex),
    MakeEntry(81, "setgroups", kInt, kHex),
    MakeEntry(83, "symlink", kPath, kPath),
    MakeEntry(85, "readlink", kPath, kGen, kInt),
    MakeEntry(86, "uselib", kPath),
    MakeEntry(87, "swapon", kHex, kHex, kHex, kHex),
    MakeEntry(88, "reboot", kGen, kGen, kGen, kGen),
    MakeEntry(91, "munmap", kHex, kHex),
    MakeEntry(92, "truncate", kPath, kHex, kHex, kHex),
    MakeEntry(93, "ftruncate", kInt, kInt),
    MakeEntry(94, "fchmod", kInt, kOct),
    MakeEntry(95, "fchown", kInt, kInt, kInt),
    MakeEntry(96, "getpriority", kInt, kInt),
    MakeEntry(97, "setpriority", kInt, kInt, kInt),
    MakeEntry(99, "statfs", kPath, kHex),
    MakeEntry(100, "fstatfs", kInt, kHex),
    MakeEntry(103, "syslog", kInt, kHex, kInt),
    MakeEntry(104, "setitimer", kInt, kHex, kHex),
    MakeEntry(105, "getitimer", kInt, kHex),
    MakeEntry(106, "stat", kPath, kGen),
    MakeEntry(107, "lstat", kPath, kGen),
    MakeEntry(108, "fstat", kHex, kHex, kHex, kHex),
    MakeEntry(111, "vhangup"),
    MakeEntry(114, "wait4", kHex, kHex, kHex, kHex),
    MakeEntry(115, "swapoff", kHex, kHex, kHex, kHex),
    MakeEntry(116, "sysinfo", kHex),
    MakeEntry(118, "fsync", kInt),
    MakeEntry(119, "sigreturn", kHex, kHex, kHex, kHex),
    MakeEntry(120, "clone", kCloneFlag, kHex, kHex, kHex, kHex),
    MakeEntry(121, "setdomainname", kString, kInt),
    MakeEntry(122, "uname", kInt),
    MakeEntry(124, "adjtimex", kHex),
    MakeEntry(125, "mprotect", kHex, kHex, kHex),
    MakeEntry(126, "sigprocmask", kHex, kHex, kHex, kHex),
    MakeEntry(128, "init_module", kGen, kInt, kString),
    MakeEntry(129, "delete_module", kString, kHex),
    MakeEntry(131, "quotactl", kHex, kHex, kHex, kHex),
    MakeEntry(132, "getpgid", kInt),
    MakeEntry(133, "fchdir", kInt),
    MakeEntry(134, "bdflush", kHex, kHex, kHex, kHex),
    MakeEntry(135, "sysfs", kInt, kInt, kInt),
    MakeEntry(136, "personality", kHex),
    MakeEntry(138, "setfsuid", kInt),
    MakeEntry(139, "setfsgid", kInt),
    MakeEntry(140, "_llseek", kHex, kHex, kHex, kHex),
    MakeEntry(141, "getdents", kInt, kHex, kInt),
    MakeEntry(142, "_newselect", kHex, kHex, kHex, kHex),
    MakeEntry(143, "flock", kInt, kInt),
    MakeEntry(144, "msync", kHex, kInt, kInt),
    MakeEntry(145, "readv", kInt, kHex, kInt),
    MakeEntry(146, "writev", kInt, kHex, kInt),
    MakeEntry(147, "getsid", kInt),
    MakeEntry(148, "fdatasync", kInt),
    MakeEntry(149, "_sysctl", kHex),
    MakeEntry(150, "mlock", kInt, kInt),
    MakeEntry(151, "munlock", kInt, kInt),
    MakeEntry(152, "mlockall", kHex),
    MakeEntry(153, "munlockall"),
    MakeEntry(154, "sched_setparam", kInt, kHex),
    MakeEntry(155, "sched_getparam", kInt, kHex),
    MakeEntry(156, "sched_setscheduler", kInt, kInt, kHex),
    MakeEntry(157, "sched_getscheduler", kInt),
    MakeEntry(158, "sched_yield"),
    MakeEntry(159, "sched_get_priority_max", kHex),
    MakeEntry(160, "sched_get_priority_min", kHex),
    MakeEntry(161, "sched_rr_get_interval", kInt, kHex),
    MakeEntry(162, "nanosleep", kHex, kHex),
    MakeEntry(163, "mremap", kHex, kInt, kInt, kInt, kHex),
    MakeEntry(164, "setresuid", kInt, kInt, kInt),
    MakeEntry(165, "getresuid", kHex, kHex, kHex),
    MakeEntry(168, "poll", kGen, kInt, kInt),
    MakeEntry(169, "nfsservctl", kInt, kGen, kGen),
    MakeEntry(170, "setresgid", kInt, kInt, kInt),
    MakeEntry(171, "getresgid", kHex, kHex, kHex),
    MakeEntry(172, "prctl", kInt, kHex, kHex, kHex, kHex),
    MakeEntry(173, "rt_sigreturn"),
    MakeEntry(174, "rt_sigaction", kHex, kHex, kHex, kHex),
    MakeEntry(175, "rt_sigprocmask", kGen, kGen, kGen, kGen),
    MakeEntry(176, "rt_sigpending", kHex, kInt),
    MakeEntry(177, "rt_sigtimedwait", kGen, kGen, kGen, kGen),
    MakeEntry(178, "rt_sigqueueinfo", kInt, kSignal, kHex),
    MakeEntry(179, "rt_sigsuspend", kHex, kInt),
    MakeEntry(180, "pread64", kGen, kGen, kGen, kGen),
    MakeEntry(181, "pwrite64", kGen, kGen, kGen, kGen),
    MakeEntry(182, "chown", kHex, kHex, kHex, kHex),
    MakeEntry(183, "getcwd", kHex, kInt),
    MakeEntry(184, "capget", kHex, kHex),
    MakeEntry(185, "capset", kHex, kHex),
    MakeEntry(186, "sigaltstack", kHex, kHex),
    MakeEntry(187, "sendfile", kGen, kGen, kGen, kGen),
    MakeEntry(190, "vfork"),
    MakeEntry(191, "ugetrlimit", kHex, kHex, kHex, kHex),
    MakeEntry(192, "mmap2", kHex, kHex, kHex, kHex),
    MakeEntry(193, "truncate64", kHex, kHex, kHex, kHex),
    MakeEntry(194, "ftruncate64", kHex, kHex, kHex, kHex),
    MakeEntry(195, "stat64", kHex, kHex, kHex, kHex),
    MakeEntry(196, "lstat64", kHex, kHex, kHex, kHex),
    MakeEntry(197, "fstat64", kHex, kHex, kHex, kHex),
    MakeEntry(198, "lchown32", kHex, kHex, kHex, kHex),
    MakeEntry(199, "getuid32", kHex, kHex, kHex, kHex),
    MakeEntry(200, "getgid32", kHex, kHex, kHex, kHex),
    MakeEntry(201, "geteuid32", kHex, kHex, kHex, kHex),
    MakeEntry(202, "getegid32", kHex, kHex, kHex, kHex),
    MakeEntry(203, "setreuid32", kHex, kHex, kHex, kHex),
    MakeEntry(204, "setregid32", kHex, kHex, kHex, kHex),
    MakeEntry(205, "getgroups32", kHex, kHex, kHex, kHex),
    MakeEntry(206, "setgroups32", kHex, kHex, kHex, kHex),
    MakeEntry(207, "fchown32", kHex, kHex, kHex, kHex),
    MakeEntry(208, "setresuid32", kHex, kHex, kHex, kHex),
    MakeEntry(209, "getresuid32", kHex, kHex, kHex, kHex),
    MakeEntry(210, "setresgid32", kHex, kHex, kHex, kHex),
    MakeEntry(211, "getresgid32", kHex, kHex, kHex, kHex),
    MakeEntry(212, "chown32", kHex, kHex, kHex, kHex),
    MakeEntry(213, "setuid32", kHex, kHex, kHex, kHex),
    MakeEntry(214, "setgid32", kHex, kHex, kHex, kHex),
    MakeEntry(215, "setfsuid32", kHex, kHex, kHex, kHex),
    MakeEntry(216, "setfsgid32", kHex, kHex, kHex, kHex),
    MakeEntry(217, "getdents64", kInt, kHex, kInt),
    MakeEntry(218, "pivot_root", kHex, kHex, kHex, kHex),
    MakeEntry(219, "mincore", kHex, kInt, kHex),
    MakeEntry(220, "madvise", kHex, kInt, kInt),
    MakeEntry(221, "fcntl64", kHex, kHex, kHex, kHex),
    MakeEntry(224, "gettid"),
    MakeEntry(225, "readahead", kInt, kInt, kInt),
    MakeEntry(226, "setxattr", kPath, kString, kGen, kInt, kHex),
    MakeEntry(227, "lsetxattr", kPath, kString, kGen, kInt, kHex),
    MakeEntry(228, "fsetxattr", kInt, kString, kGen, kInt, kHex),
    MakeEntry(229, "getxattr", kHex, kHex, kHex, kHex),
    MakeEntry(230, "lgetxattr", kHex, kHex, kHex, kHex),
    MakeEntry(231, "fgetxattr", kGen, kGen, kGen, kGen),
    MakeEntry(232, "listxattr", kHex, kHex, kHex, kHex),
    MakeEntry(233, "llistxattr", kHex, kHex, kHex, kHex),
    MakeEntry(234, "flistxattr", kInt, kGen, kInt),
    MakeEntry(235, "removexattr", kHex, kHex, kHex, kHex),
    MakeEntry(236, "lremovexattr", kPath, kString),
    MakeEntry(237, "fremovexattr", kInt, kString),
    MakeEntry(238, "tkill", kHex, kHex, kHex, kHex),
    MakeEntry(239, "sendfile64", kHex, kHex, kHex, kHex),
    MakeEntry(240, "futex", kGen, kInt, kInt, kGen, kGen, kInt),
    MakeEntry(241, "sched_setaffinity", kInt, kInt, kHex),
    MakeEntry(242, "sched_getaffinity", kInt, kInt, kHex),
    MakeEntry(243, "io_setup", kInt, kHex),
    MakeEntry(244, "io_destroy", kInt),
    MakeEntry(245, "io_getevents", kInt, kInt, kInt, kHex, kHex),
    MakeEntry(246, "io_submit", kInt, kInt, kHex),
    MakeEntry(247, "io_cancel", kInt, kHex, kHex),
    MakeEntry(248, "exit_group", kHex, kHex, kHex, kHex),
    MakeEntry(249, "lookup_dcookie", kInt, kString, kInt),
    MakeEntry(250, "epoll_create", kInt),
    MakeEntry(251, "epoll_ctl", kGen, kGen, kGen, kGen),
    MakeEntry(252, "epoll_wait", kGen, kGen, kGen, kGen),
    MakeEntry(253, "remap_file_pages", kGen, kInt, kInt, kInt, kHex),
    MakeEntry(256, "set_tid_address", kHex),
    MakeEntry(257, "timer_create", kInt, kHex, kHex),
    MakeEntry(258, "timer_settime", kGen, kGen, kGen, kGen),
    MakeEntry(259, "timer_gettime", kInt, kHex),
    MakeEntry(260, "timer_getoverrun", kInt),
    MakeEntry(261, "timer_delete", kInt),
    MakeEntry(262, "clock_settime", kInt, kHex),
    MakeEntry(263, "clock_gettime", kInt, kHex),
    MakeEntry(264, "clock_getres", kInt, kHex),
    MakeEntry(265, "clock_nanosleep", kGen, kGen, kGen, kGen),
    MakeEntry(266, "statfs64", kHex, kHex, kHex, kHex),
    MakeEntry(267, "fstatfs64", kHex, kHex, kHex, kHex),
    MakeEntry(268, "tgkill", kHex, kHex, kHex, kHex),
    MakeEntry(269, "utimes", kPath, kHex),
    MakeEntry(271, "pciconfig_iobase", kHex, kHex, kHex, kHex),
    MakeEntry(272, "pciconfig_read", kHex, kHex, kHex, kHex),
    MakeEntry(273, "pciconfig_write", kHex, kHex, kHex, kHex),
    MakeEntry(274, "mq_open", kGen, kGen, kGen, kGen),
    MakeEntry(275, "mq_unlink", kString),
    MakeEntry(276, "mq_timedsend", kHex, kHex, kInt, kInt, kHex),
    MakeEntry(277, "mq_timedreceive", kHex, kHex, kInt, kHex, kHex),
    MakeEntry(278, "mq_notify", kHex, kHex),
    MakeEntry(279, "mq_getsetattr", kHex, kHex, kHex),
    MakeEntry(280, "waitid", kInt, kInt, kHex, kInt, kHex),
    MakeEntry(281, "socket", kAddressFamily, kInt, kInt),
    MakeEntry(282, "bind", kInt, kSockaddr, kInt),
    MakeEntry(283, "connect", kInt, kSockaddr, kInt),
    MakeEntry(284, "listen", kInt, kInt),
    MakeEntry(285, "accept", kInt, kSockaddr, kHex),
    MakeEntry(286, "getsockname", kInt, kSockaddr, kHex),
    MakeEntry(287, "getpeername", kInt, kSockaddr, kHex),
    MakeEntry(288, "socketpair", kGen, kGen, kGen, kGen),
    MakeEntry(289, "send", kHex, kHex, kHex, kHex),
    MakeEntry(290, "sendto", kInt, kGen, kInt, kHex, kSockaddr, kInt),
    MakeEntry(291, "recv", kHex, kHex, kHex, kHex),
    MakeEntry(292, "recvfrom", kInt, kHex, kInt, kHex, kSockaddr, kHex),
    MakeEntry(293, "shutdown", kInt, kInt),
    MakeEntry(294, "setsockopt", kInt, kInt, kInt, kHex, kHex),
    MakeEntry(295, "getsockopt", kInt, kInt, kInt, kHex, kInt),
    MakeEntry(296, "sendmsg", kInt, kSockmsghdr, kHex),
    MakeEntry(297, "recvmsg", kInt, kHex, kInt),
    MakeEntry(298, "semop", kInt, kHex, kInt),
    MakeEntry(299, "semget", kInt, kInt, kHex),
    MakeEntry(300, "semctl", kInt, kInt, kInt, kHex),
    MakeEntry(301, "msgsnd", kInt, kHex, kInt, kHex),
    MakeEntry(302, "msgrcv", kInt, kHex, kInt, kInt, kHex),
    MakeEntry(303, "msgget", kInt, kHex),
    MakeEntry(304, "msgctl", kInt, kInt, kHex),
    MakeEntry(305, "shmat", kInt, kHex, kHex),
    MakeEntry(306, "shmdt", kHex),
    MakeEntry(307, "shmget", kInt, kInt, kHex),
    MakeEntry(308, "shmctl", kInt, kInt, kHex),
    MakeEntry(309, "add_key", kString, kString, kGen, kInt, kInt),
    MakeEntry(310, "request_key", kGen, kGen, kGen, kGen),
    MakeEntry(311, "keyctl", kInt, kInt, kInt, kInt, kInt),
    MakeEntry(312, "semtimedop", kInt, kHex, kInt, kHex),
    MakeEntry(313, "vserver", kHex, kHex, kHex, kHex),
    MakeEntry(314, "ioprio_set", kInt, kInt, kInt),
    MakeEntry(315, "ioprio_get", kInt, kInt),
    MakeEntry(316, "inotify_init"),
    MakeEntry(317, "inotify_add_watch", kInt, kPath, kHex),
    MakeEntry(318, "inotify_rm_watch", kInt, kInt),
    MakeEntry(319, "mbind", kGen, kInt, kInt, kHex, kInt, kHex),
    MakeEntry(320, "get_mempolicy", kInt, kHex, kInt, kInt, kHex),
    MakeEntry(321, "set_mempolicy", kInt, kHex, kInt),
    MakeEntry(322, "openat", kGen, kPath, kOct, kHex, kGen, kGen),
    MakeEntry(323, "mkdirat", kInt, kPath, kOct),
    MakeEntry(324, "mknodat", kInt, kPath, kOct, kHex),
    MakeEntry(325, "fchownat", kInt, kPath, kInt, kInt, kHex),
    MakeEntry(326, "futimesat", kInt, kPath, kHex),
    MakeEntry(327, "fstatat64", kHex, kHex, kHex, kHex),
    MakeEntry(328, "unlinkat", kInt, kPath, kHex),
    MakeEntry(329, "renameat", kGen, kPath, kGen, kPath),
    MakeEntry(330, "linkat", kInt, kPath, kInt, kPath, kHex),
    MakeEntry(331, "symlinkat", kPath, kGen, kPath),
    MakeEntry(332, "readlinkat", kInt, kPath, kHex, kInt),
    MakeEntry(333, "fchmodat", kInt, kPath, kOct, kHex),
    MakeEntry(334, "faccessat", kInt, kPath, kHex, kHex),
    MakeEntry(335, "pselect6", kInt, kHex, kHex, kHex, kHex),
    MakeEntry(336, "ppoll", kHex, kInt, kHex, kHex, kInt),
    MakeEntry(337, "unshare", kHex),
    MakeEntry(338, "set_robust_list", kGen, kGen),
    MakeEntry(339, "get_robust_list", kInt, kHex, kHex),
    MakeEntry(340, "splice", kInt, kHex, kInt, kHex, kInt, kHex),
    MakeEntry(342, "tee", kGen, kGen, kGen, kGen),
    MakeEntry(343, "vmsplice", kGen, kGen, kGen, kGen),
    MakeEntry(344, "move_pages", kInt, kInt, kHex, kHex, kHex, kHex),
    MakeEntry(345, "getcpu", kHex, kHex, kHex),
    MakeEntry(346, "epoll_pwait", kInt, kHex, kInt, kInt, kHex, kInt),
    MakeEntry(347, "kexec_load", kGen, kGen, kGen, kGen),
    MakeEntry(348, "utimensat", kGen, kGen, kGen, kGen),
    MakeEntry(349, "signalfd", kInt, kHex, kHex),
    MakeEntry(350, "timerfd_create", kInt, kHex),
    MakeEntry(351, "eventfd", kInt),
    MakeEntry(352, "fallocate", kGen, kGen, kGen, kGen),
    MakeEntry(353, "timerfd_settime", kGen, kGen, kGen, kGen),
    MakeEntry(354, "timerfd_gettime", kInt, kHex),
    MakeEntry(355, "signalfd4", kGen, kGen, kGen, kGen),
    MakeEntry(356, "eventfd2", kInt, kHex),
    MakeEntry(357, "epoll_create1", kHex),
    MakeEntry(358, "dup3", kGen, kGen, kGen),
    MakeEntry(359, "pipe2", kHex, kHex),
    MakeEntry(360, "inotify_init1", kHex),
    MakeEntry(361, "preadv", kInt, kHex, kInt, kInt, kInt),
    MakeEntry(362, "pwritev", kInt, kHex, kInt, kInt, kInt),
    MakeEntry(363, "rt_tgsigqueueinfo", kGen, kGen, kGen, kGen),
    MakeEntry(364, "perf_event_open", kHex, kInt, kInt, kInt, kHex),
    MakeEntry(365, "recvmmsg", kInt, kHex, kInt, kHex, kHex),
    MakeEntry(366, "accept4", kGen, kGen, kGen, kGen),
    MakeEntry(367, "fanotify_init", kHex, kHex, kHex, kHex),
    MakeEntry(368, "fanotify_mark", kInt, kHex, kHex, kInt, kPath),
    MakeEntry(369, "prlimit64", kHex, kHex, kHex, kHex),
    MakeEntry(370, "name_to_handle_at", kInt, kPath, kHex, kHex, kHex),
    MakeEntry(371, "open_by_handle_at", kHex, kHex, kHex, kHex),
    MakeEntry(372, "clock_adjtime", kHex, kHex, kHex, kHex),
    MakeEntry(373, "syncfs", kHex, kHex, kHex, kHex),
    MakeEntry(374, "sendmmsg", kHex, kHex, kHex, kHex),
    MakeEntry(375, "setns", kHex, kHex, kHex, kHex),
    MakeEntry(376, "process_vm_readv", kInt, kHex, kInt, kHex, kInt, kInt),
    MakeEntry(377, "process_vm_writev", kInt, kHex, kInt, kHex, kInt, kInt),
    MakeEntry(378, "kcmp", kInt, kInt, kInt, kInt, kInt),
    MakeEntry(379, "finit_module", kHex, kHex, kHex, kHex),
    MakeEntry(380, "sched_setattr", kInt, kHex, kHex),
    MakeEntry(381, "sched_getattr", kGen, kGen, kGen, kGen),
    MakeEntry(382, "renameat2", kGen, kPath, kGen, kPath, kGen, kGen),
    MakeEntry(383, "seccomp", kInt, kHex, kHex),
    MakeEntry(384, "getrandom", kGen, kInt, kHex),
    MakeEntry(385, "memfd_create", kString, kHex),
    MakeEntry(386, "bpf", kHex, kHex, kHex, kHex),
    MakeEntry(387, "execveat", kInt, kPath, kHex, kHex, kHex),
    MakeEntry(388, "userfaultfd", kHex),
    MakeEntry(389, "membarrier", kHex, kHex),
    MakeEntry(390, "mlock2", kHex, kHex, kHex, kHex),
    MakeEntry(391, "copy_file_range", kInt, kHex, kInt, kHex, kInt, kHex),
    MakeEntry(392, "preadv2", kInt, kHex, kInt, kInt, kInt, kHex),
    MakeEntry(393, "pwritev2", kInt, kHex, kInt, kInt, kInt, kHex),
    MakeEntry(400, "migrate_pages", kGen, kGen, kGen, kGen),
    MakeEntry(401, "kexec_file_load", kInt, kInt, kInt, kString, kHex),
    MakeEntry(435, "clone3", kHex, kInt),
    MakeEntry(436, "close_range", kInt, kInt, kHex),
    MakeEntry(437, "openat2", kInt, kPath, kHex, kInt),
    MakeEntry(438, "pidfd_getfd", UnknownArguments()),
    MakeEntry(439, "faccessat2", kInt, kPath, kHex, kHex),
    MakeEntry(440, "process_madvise", UnknownArguments()),
    MakeEntry(441, "epoll_pwait2", UnknownArguments()),
    MakeEntry(442, "mount_setattr", UnknownArguments()),
    MakeEntry(443, "quotactl_fd", UnknownArguments()),
    MakeEntry(444, "landlock_create_ruleset", UnknownArguments()),
    MakeEntry(445, "landlock_add_rule", UnknownArguments()),
    MakeEntry(446, "landlock_restrict_self", UnknownArguments()),
    MakeEntry(448, "process_mrelease", UnknownArguments()),
    MakeEntry(449, "futex_waitv", UnknownArguments()),
    MakeEntry(450, "set_mempolicy_home_node", UnknownArguments()),
    MakeEntry(451, "cachestat", UnknownArguments()),
    MakeEntry(452, "fchmodat2", kInt, kPath, kHex, kHex),
    MakeEntry(453, "map_shadow_stack", UnknownArguments()),
    MakeEntry(454, "futex_wake", UnknownArguments()),
    MakeEntry(455, "futex_wait", UnknownArguments()),
    MakeEntry(456, "futex_requeue", UnknownArguments()),
    MakeEntry(457, "statmount", UnknownArguments()),
    MakeEntry(458, "listmount", UnknownArguments()),
    MakeEntry(459, "lsm_get_self_attr", UnknownArguments()),
    MakeEntry(460, "lsm_set_self_attr", UnknownArguments()),
    MakeEntry(461, "lsm_list_modules", UnknownArguments()),
    MakeEntry(0xf0001, "ARM_breakpoint", kHex, kHex, kHex, kHex),
    MakeEntry(0xf0002, "ARM_cacheflush", kHex, kHex, kHex, kHex),
    MakeEntry(0xf0003, "ARM_usr26", kHex, kHex, kHex, kHex),
    MakeEntry(0xf0004, "ARM_usr32", kHex, kHex, kHex, kHex),
    MakeEntry(0xf0005, "ARM_set_tls", kHex, kHex, kHex, kHex),
    // clang-format on
};

static_assert(IsSorted(kSyscallDataArm32, SyscallTable::Entry::BySyscallNr),
              "Syscalls should be sorted");

}  // namespace
}  // namespace syscalls

SyscallTable SyscallTable::get(sapi::cpu::Architecture arch) {
  switch (arch) {
    case sapi::cpu::kX8664:
      return SyscallTable(syscalls::kSyscallDataX8664);
    case sapi::cpu::kX86:
      return SyscallTable(syscalls::kSyscallDataX8632);
    case sapi::cpu::kPPC64LE:
      return SyscallTable(syscalls::kSyscallDataPPC64LE);
    case sapi::cpu::kArm64:
      return SyscallTable(syscalls::kSyscallDataArm64);
    case sapi::cpu::kArm:
      return SyscallTable(syscalls::kSyscallDataArm32);
    default:
      return SyscallTable();
  }
}

}  // namespace sandbox2
